(*
    Juli8 Monad Library for OCaml. This version from 13 April 2015.

    Mostly follows the nomenclature and implementation of the Haskell
    transformers and mtl packages. Their Reader.local method is here called
    Reader.shift. Also I used "hoist" in place of Haskell's "lift" terminology,
    as the latter is too overloaded.

    I'd guess about 5% of this code is interesting. The rest is boilerplate.
    It's ugly, repetitive code: a pain to write and I haven't been able
    to make it as clean to deploy as I'd like, either. (See below.)
    Plus I needed to resort to Obj.magic a few times to avoid making things
    even more cumbersome.

    Here's how to create a simple two-layer monad stack:

      module O = Monad.Option
      module R = Monad.Reader(struct type env = ... end)
      module OR = struct
        (* apply the OptionT transformer to the wrapped Reader monad *)
        include O.T(R)
        (* and "hoist" the Reader operations into the new monad *)
        include Monad.Derive.Reader(R)(O.T(R))
      end

    Here's how to add a third layer:

      module L = Monad.List
      module LOR = struct
        include L.T(OR)
        (* and "hoist" the Option and Reader operations from OR into the new monad; order doesn't matter *)
        include Monad.Derive.Option(OR)(L.T(OR))
        include Monad.Derive.Reader(OR)(L.T(OR))
      end

    However there are lots of variations you need to (manually) attend to:

    1. The monad transformers are typically functors named `.T` in the namespace
       of the wrapping monad. However in some cases they have a second
       letter and in some cases the first letter is not 'T`.
    2. If the wrapping monad has no `mzero` and `(++)`, but the wrapped monad has
       both, then you should use the transformer beginning with `.M` rather
       than `.T`. If the wrapped monad is providing only `mzero` (either
       because that's all it has, or because the wrapping monad already has `(++)`),
       then use `.Z`. If the wrapped monad is providing only `(++)`, then
       use `.P`.
    3. The Reader.Z* and State.Z* transformers have variants `Reader.Z'`, `Reader.Z'E`,
       and so on, which will use the underlying mzero instead of propagating `Not_found`
       from an invocation of `asks` or `gets`.
    4. If the wrapped monad contains both an Error and a Writer component,
       then the transformer should have a 'X' suffix, for example `.TX`. Also
       the names of some of the `Derive` functors will likely differ (see below).
       If the wrapped monad contains Error only, then the transformer should
       have an `E` suffix. If the wrapped monad contains Writer only, then
       the transformer should have a `W` suffix.
    5. If Deriving from a monad that contains a Writer component (and
       the components you're thereby deriving are other than Writer), then
       use `Monad.Derive.ReaderW(...)` rather than just `Monad.Derive.Reader`.
       If Deriving from a monad that contains an Error component, then
       `Monad.Derive.ReaderE(...)`. If Deriving from a monad that contains
       both, then `Monad.Derive.ReaderX(...)`. For `Derive.Error` only use the
       `W` suffix and for `Derive.Writer` only use the `E` suffix.
    6. If Deriving into a monad whose outermost wrapper is Error, append
       `_toE` to the name of the `Derive` functor, for example `Derive.Reader_toE`
       or `Derive.ReaderW_toE`. Similarly if Deriving into Writer, append
       `_toW`.
    7. If Deriving from a monad that contains (for example) a Reader component
       into a monad whose outermost wrapper is *another* Reader, then append _dbl to
       the `Derive` functor, for example `Derive.Reader_dbl` or `Derive.ReaderE_dbl`.
       Currently the library makes no provision for including more than
       one layer of Error or Writer in a monad stack.

    Examples:

      module O = Monad.Option
      module L = Monad.List
      module T = Monad.LTree
      module W = Monad.Writer1 (* a pre-configured version of Monad.Writer *)
      module E = Monad.Failure (* a pre-configured version of Monad.Error *)

      module OT = struct
        include O.P(T)
        include Monad.Derive.Tree(T)(O.P(T))
      end

      module TO = struct
        include T.Z(O)
        include Monad.Derive.Option(O)(T.Z(O))
      end

      module EOT = struct
        include E.M(OT)
        include Monad.Derive.Option_toE(OT)(E.M(OT))
        include Monad.Derive.Tree_toE(OT)(E.M(OT))
      end

      module WEOT = struct
        include W.ME(EOT)
        include Monad.Derive.Error_toW(EOT)(W.ME(EOT))
        include Monad.Derive.OptionE_toW(EOT)(W.ME(EOT))
        include Monad.Derive.TreeE_toW(EOT)(W.ME(EOT))
      end

      module LWEOT = struct
        include L.TX(WEOT)
        include Monad.Derive.WriterE(WEOT)(L.TX(WEOT))
        include Monad.Derive.ErrorW(WEOT)(L.TX(WEOT))
        include Monad.Derive.OptionX(WEOT)(L.TX(WEOT))
        include Monad.Derive.TreeX(WEOT)(L.TX(WEOT))
      end

    Other quirks:

    * hoisting List.pick currently doesn't work
    * Change docs: {Option,List}.test -> {opt,list}_map
    * haven't tried to hoist Option.opt_map, List.list_map
    * List.mids not yet provided for ListT, nor hoisted
    * Haskell exposes seq in every monad's API; however this really has to do with how the monad interacts with a wrapping ListT
    * E's result is in the OK type, but E_'s result is bare (or raise exception)

*)



module Monad = struct

  type voidlog
  type voidmsg

  module type MAPPABLE = sig
    type 'a t
    val map : ('a -> 'b) -> 'a t -> 'b t
    (* mapconst is definable as map % const. For example mapconst 4 [1,2,3] == [4,4,4]. Haskell calls mapconst <$ in Data.Functor and Control.Applicative. They also use $> for flip mapconst, and Control.Monad.void for mapconst (). *)
  end

  module type APPLICATIVE = sig
    include MAPPABLE
    val mid : 'a -> 'a t
    val map2 : ('a -> 'b -> 'c) -> 'a t -> 'b t -> 'c t
    val mapply : ('a -> 'b) t -> 'a t -> 'b t
    val (>>) : 'a t -> 'b t -> 'b t
    val (<<) : 'a t -> 'b t -> 'a t
  end

  module type MONADBASE = sig
    include APPLICATIVE
    val (>>=) : 'a t -> ('a -> 'b t) -> 'b t
    val (>=>) : ('a -> 'b t) -> ('b -> 'c t) -> ('a -> 'c t)
    val (<=<) : ('b -> 'c t) -> ('a -> 'b t) -> ('a -> 'c t)
    val join : 'a t t -> 'a t
    val ignore : 'a t -> unit t
    val seq : 'a t list -> 'a list t
    val seq_ignore : unit t list -> unit t
    val do_when : bool -> unit t -> unit t
    val do_unless : bool -> unit t -> unit t
  end

  module type MONAD = sig
    include MONADBASE
    type 'a result
    val run : 'a t -> 'a result
  end

  module type MONADT = sig
    include MONADBASE
    type 'a result
    val run : 'a t -> 'a result
    type 'a ut
    type 'a output
    type ulog
    type umsg
    val hoist : 'a ut -> 'a t
    val hoist_map : ('a output ut -> 'b output ut) -> 'a t -> 'b t
    val hoist_pick : ('a output ut -> ('a output * 'a output ut) ut) -> 'a t -> ('a * 'a t) t
    val hoist_listen : ('a output ut -> ('a output * ulog) ut) -> 'a t -> ('a * ulog) t
    val hoist_pass : (('a output * (ulog -> ulog)) ut -> 'a output ut) -> ('a * (ulog -> ulog)) t -> 'a t
    val hoist_catch : ('a output ut -> (umsg -> 'a output ut) -> 'a output ut) -> 'a t -> (umsg -> 'a t) -> 'a t
    (*
      val hmap : ('a ut -> 'b ut) -> 'a t -> 'b t
      let hmap f xx = xx >>= hoist % f % U.mid
      -- List
      let hmap f xx = U.(xx >>= distribute (f % mid)) (* more efficient? *)
      -- Tree
      let hmap f xx = U.(xx >>= distribute (f % mid)) (* more efficient? *)
      -- Reader
      (* was using hmap f xx = fun e -> f (xx e), but that gives different results with underlying List than official hmap *)
      let hmap f xx = fun e -> U.(xx e >>= f % mid) (* more efficient *)
      -- Writer
      let hmap f xx = U.(xx >>= fun (x,w) -> f (mid x) >>= fun z -> mid (z,w)) (* more efficient *)
    *)
    (*
      -- Option
      let unhoisted2 f xx yy = U.(xx >>= function None -> mid None | Some x -> yy >>= function None -> mid None | Some y -> f (mid x) (mid y) >>= fun z -> mid (Some z))
      -- Reader
      let unhoisted2 (f : 'a U.t -> 'b U.t -> 'c U.t) (xx : 'd t) (yy : 'e t) : 'f t =
        fun e -> f (xx e) (yy e)
      -- State
      (* note that the use of U.(>>=) in implementing ++ will cross underlying lists *)
      let unhoisted2 (f : 'a U.t -> 'b U.t -> 'c U.t) (xx : 'd t) (yy : 'e t) : 'f t =
        fun s -> U.(xx s >>= fun (x, s') -> yy s' >>= fun (y, s'') -> f (mid x) (mid y) >>= fun z -> mid (z, s''))
      -- Ref
      (* note that the use of U.(>>=) in implementing ++ will cross underlying lists *)
      let unhoisted2 (f : 'a U.t -> 'b U.t -> 'c U.t) (xx : 'd t) (yy : 'e t) : 'f t =
        fun s -> U.(xx s >>= fun (x, s') -> yy s' >>= fun (y, s'') -> f (mid x) (mid y) >>= fun z -> mid (z, s''))
      -- Writer
      (* Using unhoisted2 U.(++) to implement ++ would (i) cross underlying lists, plus (2) give each element in the result the concatenated log of all elements. *)
      let unhoisted2 (f : 'a U.t -> 'b U.t -> 'c U.t) (xx : 'a t) (yy: 'b t) : 'c t =
        U.(xx >>= fun (x,w) -> yy >>= fun (y,w') -> f (mid x) (mid y) >>= fun z -> U.mid (z, W.append w w'))
      -- Error
      let unhoisted2 (f : 'a U.t -> 'b U.t -> 'c U.t) (xx : 'a t) (yy: 'b t) : 'c t =
        U.(xx >>= function Error e -> mid (Error e) | OK x -> yy >>= function Error e -> mid (Error e) | OK y -> f (mid x) (mid y) >>= fun z -> mid (OK z))
*)
  end

  module type MONADTW = sig
    type log
    include MONADT with type ulog := log
  end

  module type MONADTE = sig
    type msg
    include MONADT with type umsg := msg
  end

  module type MONADTX = sig
    type log
    type msg
    include MONADT with type ulog := log and type umsg := msg
  end

  module type ZERO = sig
    type 'a t
    (* mzero is a value of type α that is exemplified by Nothing for the box type Maybe α and by [] for the box type List α. It has the behavior that anything ¢ mzero == mzero == mzero ¢ anything == mzero >>= anything. In Haskell, this notion is called Control.Applicative.empty or Control.Monad.mzero. *)
    val mzero : 'a t
    val guard : bool -> unit t
  end

  module type PLUS = sig
    type 'a t
    (* we don't require mzero, nor that this operation be associative *)
    val (++) : 'a t -> 'a t -> 'a t (* monadically append *)
  end

  module type MONADZERO = sig
    include MONAD
    include ZERO with type 'a t := 'a t
  end

  module type MONADPLUS = sig
    include MONAD
    include PLUS with type 'a t := 'a t
  end

  module type MONADZEROPLUS = sig
    include MONAD
    include ZERO with type 'a t := 'a t
    include PLUS with type 'a t := 'a t
  end

  module type MONADW = sig
    include MONAD
    type log
  end

  module type MONADE = sig
    include MONAD
    type msg
  end

  module type MONADX = sig
    include MONAD
    type log
    type msg
  end

  module type MONADZEROW = sig
    include MONADZERO
    type log
  end

  module type MONADZEROE = sig
    include MONADZERO
    type msg
  end

  module type MONADZEROX = sig
    include MONADZERO
    type log
    type msg
  end

  module type MONADPLUSW = sig
    include MONADPLUS
    type log
  end

  module type MONADPLUSE = sig
    include MONADPLUS
    type msg
  end

  module type MONADPLUSX = sig
    include MONADPLUS
    type log
    type msg
  end

  module type MONADZEROPLUSW = sig
    include MONADZEROPLUS
    type log
  end

  module type MONADZEROPLUSE = sig
    include MONADZEROPLUS
    type msg
  end

  module type MONADZEROPLUSX = sig
    include MONADZEROPLUS
    type log
    type msg
  end

  module type MONADZEROT = sig
    include MONADT
    include ZERO with type 'a t := 'a t
  end

  module type MONADPLUST = sig
    include MONADT
    include PLUS with type 'a t := 'a t
  end

  module type MONADZEROPLUST = sig
    include MONADT
    include ZERO with type 'a t := 'a t
    include PLUS with type 'a t := 'a t
  end


  module type OPTIONBASE = sig
    type 'a t
  end

  module type OPTION = sig
    include OPTIONBASE
    include MONADZERO with type 'a t := 'a t and type 'a result = 'a option
    val opt_map : ('a option -> 'b option) -> 'a t -> 'b t
  end

  module type OPTION_ = sig
    include OPTIONBASE
    include MONADZEROT with type 'a t := 'a t
  end

  module type OPTIONT = sig
    type 'a uresult
    include OPTION_ with type 'a result = 'a option uresult
    val opt_map : ('a option -> 'b option) -> 'a t -> 'b t
  end

  module type OPTIONPLUST = sig
    include OPTIONBASE
    type 'a uresult
    include MONADZEROPLUST with type 'a t := 'a t and type 'a result = 'a option uresult
    val opt_map : ('a option -> 'b option) -> 'a t -> 'b t
  end

  module type LISTBASE = sig
    type 'a t
    val pick : 'a t -> ('a * 'a t) t (* monadically pick each element *)
  end

  module type LIST = sig
    include LISTBASE
    include MONADZEROPLUS with type 'a t := 'a t and type 'a result = 'a list
    val list_map : ('a list -> 'b list) -> 'a t -> 'b t
  end

  module type LIST_ = sig
    include LISTBASE
    include MONADZEROPLUST with type 'a t := 'a t
    val list_map : ('a list -> 'b list) -> 'a t -> 'b t
    (*
        Monadically seq k over box<a>.
        OptionM.seq (List.map (\a -> OptionM.mid $ a+1) int_list) == (after running)
        ListOption.distribute (\a -> OptionM.mid $ a+1) int_list == Some [x+1,x+1,...]
        TreeOption.distribute (\a -> OptionM.mid $ a+1) int_tree: works similarly
    *)
    val distribute : ('a -> 'b ut) -> 'a list -> 'b t
  end

  module type LISTT = sig
    type 'a uresult
    include LIST_ with type 'a result = 'a list uresult
  end

  module type TREEBASE = sig
    type 'a t
    type 'a tree
    val branch : 'a tree -> 'a tree -> 'a tree
    val leaf : 'a -> 'a tree
    val traverse : ('b -> 'b -> 'b) -> ('a -> 'b) -> 'a tree -> 'b
  end

  module type TREE = sig
    include TREEBASE
    include MONADPLUS with type 'a t := 'a t and type 'a result = 'a tree
  end

  module type TREE_ = sig
    include TREEBASE
    include MONADPLUST with type 'a t := 'a t
    (*
        Monadically seq k over box<a>.
        OptionM.seq (List.map (\a -> OptionM.mid $ a+1) int_list) == (after running)
        ListOption.distribute (\a -> OptionM.mid $ a+1) int_list == Some [x+1,x+1,...]
        TreeOption.distribute (\a -> OptionM.mid $ a+1) int_tree: works similarly
    *)
    val distribute : ('a -> 'b ut) -> 'a tree -> 'b t
  end

  module type TREET = sig
    type 'a tree
    type 'a uresult
    include TREE_ with type 'a tree := 'a tree and type 'a result = 'a tree uresult
  end

  module type READERBASE = sig
    type 'a t
    type env
    val ask : env t
    val asks : (env -> 'a) -> 'a t
    val shift : (env -> env) -> 'a t -> 'a t
  end

  module type READER2BASE = sig
    type 'a t
    type env
    val ask : env t
    val asks : (env -> 'a) -> 'a t
    val shift : (env -> env) -> 'a t -> 'a t
    type uenv
    val uask : uenv t
    val uasks : (uenv -> 'a) -> 'a t
    val ushift : (uenv -> uenv) -> 'a t -> 'a t
  end

  module type READER = sig
    include READERBASE
    include MONAD with type 'a t := 'a t and type 'a result = env -> 'a
  end

  module type READER_ = sig
    include READERBASE
    include MONADT with type 'a t := 'a t
  end

  module type READERT = sig
    type env
    type 'a uresult
    include READER_ with type env := env and type 'a result = env -> 'a uresult
  end

  module type STATEBASE = sig
    type 'a t
    type store
    val get : store t
    val gets : (store -> 'a) -> 'a t
    val put : store -> unit t
    val modify : (store -> store) -> unit t
  end

  module type STATE = sig
    include STATEBASE
    include MONAD with type 'a t := 'a t and type 'a result = store -> 'a * store
  end

  module type STATE_ = sig
    include STATEBASE
    include MONADT with type 'a t := 'a t
  end

  module type STATET = sig
    type 'a uresult
    type store
    include STATE_ with type store := store and type 'a result = store -> ('a * store) uresult
  end

  module type REFBASE = sig
    type 'a t
    type ref
    type value
    val newref : value -> ref t
    val getref : ref -> value t
    val putref : ref -> value -> unit t
  end

  module type REF = sig
    include REFBASE
    include MONAD with type 'a t := 'a t and type 'a result = 'a
  end

  module type REF_ = sig
    include REFBASE
    include MONADT with type 'a t := 'a t
  end

  module type REFT = sig
    type 'a uresult
    include REF_ with type 'a result = 'a uresult
  end

  module type WRITERBASE = sig
    type 'a t
    type log
    (*
    listen xx : execute xx and add its log to the payload as well as logging it
    listens f xx : as listens but only add filtered log to payload
    censor f xx : execute xx while filtering its log (payload unaffected)
    *)
    val listen : 'a t -> ('a * log) t
    val listens : (log -> 'b) -> 'a t -> ('a * 'b) t
    val tell : log -> unit t
    val pass : ('a * (log -> log)) t -> 'a t
    val censor : (log -> log) -> 'a t -> 'a t
  end

  module type WRITER = sig
    include WRITERBASE
    include MONAD with type 'a t := 'a t and type 'a result = 'a * log
  end

  module type WRITER_ = sig
    include WRITERBASE
    include MONADT with type 'a t := 'a t and type ulog := log
  end

  module type WRITERT = sig
    type log
    type 'a uresult
    include WRITER_ with type log := log and type 'a result = ('a * log) uresult
  end

  module type ERRORBASE = sig
    type 'a t
    type msg
    type 'a error
    val throw : msg -> 'a t
    val catch : 'a t -> (msg -> 'a t) -> 'a t
  end

  module type ERROR = sig
    include ERRORBASE
    include MONAD with type 'a t := 'a t and type 'a result = 'a error
  end

  module type ERROR_ = sig
    include ERRORBASE
    include MONADT with type 'a t := 'a t and type umsg := msg
  end

  module type ERRORT = sig
    type 'a uresult
    include ERROR_ with type 'a result = 'a uresult (* note the difference from ERROR *)
  end


  module Make = struct

    module type MAP2 = sig
      type 'a t
      val mid : 'a -> 'a t
      val map2 : ('a -> 'b -> 'c) -> 'a t -> 'b t -> 'c t
      val map : [`Generate | `Custom of ('a -> 'b) -> 'a t -> 'b t]
      val mapply : [`Generate | `Custom of ('a -> 'b) t -> 'a t -> 'b t]
    end

    module type MAPPLY = sig
      type 'a t
      val mid : 'a -> 'a t
      val mapply : ('a -> 'b) t -> 'a t -> 'b t
      val map : [`Generate | `Custom of ('a -> 'b) -> 'a t -> 'b t]
      val map2 : [`Generate | `Custom of ('a -> 'b -> 'c) -> 'a t -> 'b t -> 'c t]
    end

    module type BIND = sig
      type 'a t
      type 'a result
      val run : 'a t -> 'a result
      val mid : 'a -> 'a t
      val (>>=) : 'a t -> ('a -> 'b t) -> 'b t
      val map : [`Generate | `Custom of ('a -> 'b) -> 'a t -> 'b t]
      val map2 : [`Generate | `Custom of ('a -> 'b -> 'c) -> 'a t -> 'b t -> 'c t]
      val mapply : [`Generate | `Custom of ('a -> 'b) t -> 'a t -> 'b t]
    end

    module type COMP = sig
      type 'a t
      type 'a result
      val run : 'a t -> 'a result
      val mid : 'a -> 'a t
      val (>=>) : ('a -> 'b t) -> ('b -> 'c t) -> ('a -> 'c t)
      val map : [`Generate | `Custom of ('a -> 'b) -> 'a t -> 'b t]
      val map2 : [`Generate | `Custom of ('a -> 'b -> 'c) -> 'a t -> 'b t -> 'c t]
      val mapply : [`Generate | `Custom of ('a -> 'b) t -> 'a t -> 'b t]
    end

    module type JOIN = sig
      type 'a t
      type 'a result
      val run : 'a t -> 'a result
      val mid : 'a -> 'a t
      val join : 'a t t -> 'a t
      val map : ('a -> 'b) -> 'a t -> 'b t
      val map2 : [`Generate | `Custom of ('a -> 'b -> 'c) -> 'a t -> 'b t -> 'c t]
      val mapply : [`Generate | `Custom of ('a -> 'b) t -> 'a t -> 'b t]
    end

    module type TRANSBASE = sig
      type 'a t
      type 'a ut
      type ulog
      type umsg
      type 'a output
      type 'a result
      val run : 'a t -> 'a result
      val (>>=) : 'a t -> ('a -> 'b t) -> 'b t
      val hoist_map : [`Generate | `Custom of ('a output ut -> 'b output ut) -> 'a t -> 'b t]
      val hoist_listen : [`Generate | `Custom of ('a output ut -> ('a output * ulog) ut) -> 'a t -> ('a * ulog) t ]
      val hoist_pass : [`Generate | `Custom of (('a output * (ulog -> ulog)) ut -> 'a output ut) -> ('a * (ulog -> ulog)) t -> 'a t ]
      val hoist_catch : [`Generate | `Custom of ('a output ut -> (umsg -> 'a output ut) -> 'a output ut) -> 'a t -> (umsg -> 'a t) -> 'a t ]
      val hoist_pick : [`Generate | `Custom of ('a output ut -> ('a output * 'a output ut) ut) -> 'a t -> ('a * 'a t) t ]
    end

    module type TRANS = sig
      module U : MONAD
      include TRANSBASE with type 'a ut := 'a U.t
      (* Provide hoist, >>=; LAWS: 1. hoist U.(mid x) == mid x; 2. hoist U.(uu >>= k) == hoist uu >>= fun u -> hoist (k u) *)
      val hoist : 'a U.t -> 'a t
    end

    module type TRANSUZ = sig
      module U : MONADZERO
      include TRANSBASE with type 'a ut := 'a U.t
      val hoist : 'a U.t -> 'a t
    end

    module type TRANSZ = sig
      module U : MONAD
      include TRANSBASE with type 'a ut := 'a U.t
      val hoist : 'a U.t -> 'a t
      val mzero : 'a t
    end

    module ApplicativeFromBind (B : BIND) : APPLICATIVE with type 'a t = 'a B.t = struct
      type 'a t = 'a B.t
      let mid = B.mid
      let (>>=) = B.(>>=)
      let map = match B.map with
      | `Custom map -> map
      | `Generate -> fun f xx -> xx >>= fun x -> mid (f x)
      let map2 = match B.map2 with
      | `Custom map2 -> map2
      | `Generate -> fun f xx yy -> xx >>= fun x -> yy >>= fun y -> mid (f x y)
      let mapply = match B.map2 with
      | `Custom map2 -> fun eta -> map2 ident eta
      | `Generate -> fun ff xx -> ff >>= fun f -> map f xx
      let (>>) xx yy = xx >>= fun _ -> yy
      let (<<) xx yy = mapply (map const xx) yy
    end

    module ApplicativeFromMap2 (B : MAP2) : APPLICATIVE with type 'a t = 'a B.t = struct
      type 'a t = 'a B.t
      let mid = B.mid
      let map2 = B.map2
      let mapply = match B.mapply with
      | `Custom mapply -> mapply
      | `Generate -> fun eta -> map2 ident eta
      let map = match B.map with
      | `Custom map -> map
      | `Generate -> fun f xx -> mapply (mid f) xx
      let (>>) xx yy = mapply (map (const ident) xx) yy
      let (<<) xx yy = mapply (map const xx) yy
    end

    module ApplicativeFromApply (B : MAPPLY) : APPLICATIVE with type 'a t = 'a B.t = struct
      type 'a t = 'a B.t
      let mid = B.mid
      let mapply = B.mapply
      let map = match B.map with
      | `Custom map -> map
      | `Generate -> fun f xx -> mapply (mid f) xx
      let map2 = match B.map2 with
      | `Custom map2 -> map2
      | `Generate -> fun f xx yy -> mapply (map f xx) yy
      let (>>) xx yy = mapply (map (const ident) xx) yy
      let (<<) xx yy = mapply (map const xx) yy
    end

    module MonadFromBind (B : BIND) : MONAD with type 'a t = 'a B.t and type 'a result = 'a B.result = struct
      let (>>=) = B.(>>=)
      include ApplicativeFromBind(B)
      type 'a result = 'a B.result
      let run = B.run
      let (>=>) j k = fun a -> j a >>= k
      let (<=<) k j = fun a -> j a >>= k
      let join xxx = xxx >>= ident
      let ignore xx = map (fun _ -> ()) xx
      (* seq xxs = let f xx zzf = (xx >>=) . flip ((zzf.).(:)) in foldr f (return $) xxs [] *)
      (*
         foldr' f z xs = foldl (\g x z -> g (f x z)) id xs z  -- foldr but evaluating from left?
         foldl'' f z xs = foldr (\x g z -> g (f z x)) id xs z -- foldl but evaluating from right? these don't work
         -- with foldr, evaluates left->right; with foldl the reverse
         seq xxs =
           let f c xx ret xs = xx >>= ret . c xs in -- careful! isn't fmap (c xs) xx because ret isn't (always) return
           reverse <$> foldr (f $ flip (:)) return xxs []
           -- or simply: foldr (f snoc) return xxs []
      *)
      let seq =
        let rec aux xs = function
        | [] -> mid (List.rev xs)
        | xx::xxs -> xx >>= fun x -> aux (x::xs) xxs in
        fun xxs -> aux [] xxs
      let rec seq_ignore = function
        | [] -> mid ()
        | xx::xxs -> xx >>= fun () -> seq_ignore xxs
      let do_when res xx = if res then xx else mid ()
      let do_unless res xx = if res then mid () else xx
    end

    module MonadFromComp (B : COMP) : MONAD with type 'a t = 'a B.t and type 'a result = 'a B.result = struct
      let (>=>) = B.(>=>)
      let (<=<) k j = j >=> k
      let (>>=) xx k = (ident >=> k) xx
      include ApplicativeFromBind(struct include B let (>>=) = (>>=) end)
      type 'a result = 'a B.result
      let run = B.run
      let join xxx = xxx >>= ident
      let ignore xx = map (fun _ -> ()) xx
      let seq =
        let rec aux xs = function
        | [] -> mid (List.rev xs)
        | xx::xxs -> xx >>= fun x -> aux (x::xs) xxs in
        fun xxs -> aux [] xxs
      let rec seq_ignore = function
        | [] -> mid ()
        | xx::xxs -> xx >>= fun () -> seq_ignore xxs
      let do_when res xx = if res then xx else mid ()
      let do_unless res xx = if res then mid () else xx
    end

    module MonadFromJoin (B : JOIN) : MONAD with type 'a t = 'a B.t and type 'a result = 'a B.result = struct
      let join = B.join
      let (>>=) xx k = join (B.map k xx)
      include ApplicativeFromBind(struct include B let (>>=) = (>>=) let map = `Custom B.map end)
      type 'a result = 'a B.result
      let run = B.run
      let (>=>) j k = fun a -> j a >>= k
      let (<=<) k j = fun a -> j a >>= k
      let ignore xx = map (fun _ -> ()) xx
      let seq =
        let rec aux xs = function
        | [] -> mid (List.rev xs)
        | xx::xxs -> xx >>= fun x -> aux (x::xs) xxs in
        fun xxs -> aux [] xxs
      let rec seq_ignore = function
        | [] -> mid ()
        | xx::xxs -> xx >>= fun () -> seq_ignore xxs
      let do_when res xx = if res then xx else mid ()
      let do_unless res xx = if res then mid () else xx
    end

    module HoistFromT(B : TRANS) = struct
      let hoist = B.hoist
      (* FIXME *)
      type 'a ut = 'a B.U.t
      type 'a output = 'a B.output
      type ulog = B.ulog
      type umsg = B.umsg
      (* Obj.magic hacks *)
      let hoist_map = match B.hoist_map with
      | `Custom hoist_map -> hoist_map
      | `Generate -> fun f xx -> Obj.magic (f (Obj.magic xx))
      let hoist_listen : ('a output ut -> ('a output * ulog) ut) -> 'a B.t -> ('a * ulog) B.t
      = match B.hoist_listen with
      | `Custom hoist_listen -> hoist_listen
      | `Generate -> fun listener xx -> hoist_map (Obj.magic listener) xx
      let hoist_pass : (('a output * (ulog -> ulog)) ut -> 'a output ut) -> ('a * (ulog -> ulog)) B.t -> 'a B.t
      = match B.hoist_pass with
      | `Custom hoist_pass -> hoist_pass
      | `Generate -> fun passer xx -> hoist_map (Obj.magic passer) xx
      let hoist_catch : ('a output ut -> (umsg -> 'a output ut) -> 'a output ut) -> 'a B.t -> (umsg -> 'a B.t) -> 'a B.t
      = match B.hoist_catch with
      | `Custom hoist_catch -> hoist_catch
      | `Generate -> fun catcher xx h -> Obj.magic (catcher (Obj.magic xx) (fun m -> Obj.magic (h m)))
      let hoist_pick : ('a output ut -> ('a output * 'a output ut) ut) -> 'a B.t -> ('a * 'a B.t) B.t
      = match B.hoist_pick with
      | `Custom hoist_pick -> hoist_pick
      | `Generate -> fun picker xx -> hoist_map (Obj.magic picker) xx

(*
      let hoist_listen listener xx = failwith "not implemented"
      let hoist_pass passer xx = failwith "not implemented"
      let hoist_catch catcher xx h = failwith "not implemented"
*)
      let hoist_pick picker xx = failwith "not implemented"
    end

    module MonadFromT (B : TRANS) : MONADT with type 'a t = 'a B.t and type 'a output := 'a B.output and type 'a result = 'a B.result and type 'a ut := 'a B.U.t and type ulog := B.ulog and type umsg := B.umsg = struct
      include MonadFromBind(struct
        include B
        let mid x = hoist U.(mid x)
        let map = `Generate let map2 = `Generate let mapply = `Generate
      end)
      include HoistFromT(B)
    end

    module MonadFromTUZ (B : TRANSUZ) : MONADZEROT with type 'a t = 'a B.t and type 'a output := 'a B.output and type 'a result = 'a B.result and type 'a ut := 'a B.U.t and type ulog := B.ulog and type umsg := B.umsg = struct
      let mzero = Obj.magic (B.hoist (B.U.mzero)) (* Obj.magic hack to generate enough polymorphism without having to thunk mzero *)
      include MonadFromBind(struct
        include B
        let (>>=) xx k = xx >>= fun x -> try k x with Match_failure _ -> mzero
        let mid x = hoist U.(mid x)
        let map = `Generate let map2 = `Generate let mapply = `Generate
      end)
      include HoistFromT(B)
      let guard res = if res then mid () else mzero
    end

    module MonadFromTZ (B : TRANSZ) : MONADZEROT with type 'a t = 'a B.t and type 'a output := 'a B.output and type 'a result = 'a B.result and type 'a ut := 'a B.U.t and type ulog := B.ulog and type umsg := B.umsg = struct
      let mzero = B.mzero
      include MonadFromBind(struct
        include B
        let (>>=) xx k = xx >>= fun x -> try k x with Match_failure _ -> mzero
        let mid x = hoist U.(mid x)
        let map = `Generate let map2 = `Generate let mapply = `Generate
      end)
      include HoistFromT(B)
      let guard res = if res then mid () else mzero
    end

  end (* Make *)

  module Option : sig
    include OPTION
    module T (U : MONAD) : OPTIONT with type 'a output = 'a option and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := voidlog and type umsg := voidmsg
    module TW (U : MONADW) : OPTIONT with type 'a output = 'a option and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := U.log and type umsg := voidmsg
    module TE (U : MONADE) : OPTIONT with type 'a output = 'a option and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := voidlog and type umsg := U.msg
    module TX (U : MONADX) : OPTIONT with type 'a output = 'a option and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := U.log and type umsg := U.msg
    module P (U : MONADPLUS) : OPTIONPLUST with type 'a output = 'a option and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := voidlog and type umsg := voidmsg
    module PW (U : MONADPLUSW) : OPTIONPLUST with type 'a output = 'a option and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := U.log and type umsg := voidmsg
    module PE (U : MONADPLUSE) : OPTIONPLUST with type 'a output = 'a option and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := voidlog and type umsg := U.msg
    module PX (U : MONADPLUSX) : OPTIONPLUST with type 'a output = 'a option and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := U.log and type umsg := U.msg
  end = struct
    include Make.MonadFromBind(struct
      type 'a t = 'a option
      type 'a result = 'a t let run xx = xx
      let map = `Custom Option.map let map2 = `Custom Option.map2 let mapply = `Generate
      let mid = some
      (* val (>>=) : 'a option -> ('a -> 'b option) -> 'b option *)
      let (>>=) xx k = match xx with Some x -> (try k x with Match_failure _ -> None) | None -> None
    end)
    let mzero = None
    let guard res : unit t = if res then Some () else None
    let opt_map f xx = f xx
    module TX (U : MONADX) = struct
      type 'a output = 'a option
      include Make.MonadFromTZ(struct
        module U = U
        type 'a t = 'a option U.t
        type 'a output = 'a option
        type 'a result = 'a option U.result let run xx = U.run xx
        type ulog = U.log
        type umsg = U.msg
        let hoist uu = U.(uu >>= fun u -> mid (Some u))
        let (>>=) xx k = U.(xx >>= function Some x -> k x | None -> mid None)
        let mzero = Obj.magic U.(mid None)
        let hoist_map = `Generate
        let hoist_listen = `Custom (fun listener -> fun xx -> U.(listener xx >>= function (None, _) -> mid None | (Some y, w) -> mid (Some (y, w))))
        let hoist_pass = `Custom (fun passer -> fun xx -> passer U.(xx >>= function None -> mid (None, ident) | Some (x, f) -> mid (Some x, f)))
        let hoist_catch = `Generate
        let hoist_pick = `Custom (fun picker -> fun xx -> U.(picker xx >>= function (None, _) -> mid None | (Some x, yy) -> mid (Some (x, yy))))
      end)
      let opt_map f xx = U.map f xx
        (* U.map (fun xo -> if p xo then xo else None) xx *)
        (* if p xx then xx else U.mid None *)
    end (* Option.TX *)
    module PX (U : MONADPLUSX) = struct
      include TX(U)
      let (++) xx yy = if yy == mzero then xx else if xx == mzero then yy else U.(xx ++ yy)
    end (* Option.PX *)
    module T (U : MONAD) = struct include TX(struct include U type log = voidlog type msg = voidmsg end) end
    module TW (U : MONADW) = struct include TX(struct include U type msg = voidmsg end) end
    module TE (U : MONADE) = struct include TX(struct include U type log = voidlog end) end
    module P (U : MONADPLUS) = struct include PX(struct include U type log = voidlog type msg = voidmsg end) end
    module PW (U : MONADPLUSW) = struct include PX(struct include U type msg = voidmsg end) end
    module PE (U : MONADPLUSE) = struct include PX(struct include U type log = voidlog end) end
  end (* Option *)


  module List : sig
    include LIST
    val mids : 'a list -> 'a t
    module T (U : MONAD) : LISTT with type 'a output = 'a list and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := voidlog and type umsg := voidmsg
    module TW (U : MONADW) : LISTT with type 'a output = 'a list and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := U.log and type umsg := voidmsg
    module TE (U : MONADE) : LISTT with type 'a output = 'a list and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := voidlog and type umsg := U.msg
    module TX (U : MONADX) : LISTT with type 'a output = 'a list and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := U.log and type umsg := U.msg
  end = struct
    include Make.MonadFromBind(struct
      type 'a t = 'a list
      type 'a result = 'a t let run xx = xx
      let map = `Custom (fun f xs -> List.map f xs) let map2 = `Custom (fun f xs -> List.map2 f xs) let mapply = `Generate
      let mid = singleton
      let (>>=) xx k = List.catmap (fun x -> try k x with Match_failure _ -> []) xx
    end)
    let mzero = []
    let guard res : unit t = if res then [()] else []
    (* (++) has tighter precedence than (>>=) *)
    let (++) = append
    let rec pick = function [] -> mzero | x::xs -> mid (x,xs) ++ (pick xs >>= fun (y,ys) -> mid (y, x::ys))
    let list_map f xx = f xx (* if p xx then xx else [] *)
    (* let mids xs = let rec aux xx = function [] -> xx | x::xs -> aux (mid x ++ xx) xs in aux mzero (rev xs) *)
    let mids xs = xs
    module TX (U : MONADX) = struct
      let distribute k xs = U.seq (List.map k xs)
      type 'a output = 'a list
      include Make.MonadFromTZ(struct
        module U = U
        type 'a t = 'a list U.t
        type 'a output = 'a list
        type 'a result = 'a list U.result let run xx = U.run xx
        type ulog = U.log
        type umsg = U.msg
        let hoist uu = U.(uu >>= fun u -> mid [u])
        let (>>=) xx k = U.(xx >>= fun xs -> distribute k xs >>= fun xss -> mid (concat xss))
        let mzero = Obj.magic U.(mid [])
        let hoist_map = `Generate
        let hoist_listen = `Custom (fun listener -> fun xx -> U.(listener xx >>= fun (xs,w) -> distribute (fun x -> mid (x,w)) xs))
        let hoist_pass = `Custom (fun passer xx -> passer U.(xx >>= fun xs -> let (fs, ys) = List.mapz (fun fs (x, f) -> (f % fs, x)) ident xs in mid (ys, fs)))
        let hoist_catch = `Generate
        let hoist_pick = `Generate
      end)
      let (++) xx yy = U.(xx >>= fun xs -> yy >>= fun ys -> mid (append xs ys))
      let rec pick xx = U.(>>=) xx (function [] -> mzero | x::xs -> mid (x, U.(mid xs)) ++ (pick U.(mid xs) >>= fun (y,yy) -> mid (y, U.(yy >>= fun ys -> mid (x::ys)))))
      let list_map f xx = U.map f xx
        (* U.map (fun xs -> if p xs then xs else []) xx *)
        (* if p xx then xx else U.mid [] *)
    end (* List.TX *)
    module T (U : MONAD) = struct include TX(struct include U type log = voidlog type msg = voidmsg end) end
    module TW (U : MONADW) = struct include TX(struct include U type msg = voidmsg end) end
    module TE (U : MONADE) = struct include TX(struct include U type log = voidlog end) end
  end (* List *)


  (* LTree, unit centers, has natural ++ *)
  (* ITree, unit leaves, has natural mzero *)

  module LTree : sig
    type 'a tree = Leaf of 'a | Branch of 'a tree * 'a tree
    include TREE with type 'a tree := 'a tree
    module T (U : MONAD) : TREET with type 'a tree = 'a tree and type 'a output = 'a tree and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := voidlog and type umsg := voidmsg
    module TW (U : MONADW) : TREET with type 'a tree = 'a tree and type 'a output = 'a tree and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := U.log and type umsg := voidmsg
    module TE (U : MONADE) : TREET with type 'a tree = 'a tree and type 'a output = 'a tree and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := voidlog and type umsg := U.msg
    module TX (U : MONADX) : TREET with type 'a tree = 'a tree and type 'a output = 'a tree and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := U.log and type umsg := U.msg
    module Z (U : MONADZERO) : sig
      include TREET with type 'a tree = 'a tree and type 'a output = 'a tree and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := voidlog and type umsg := voidmsg
      include ZERO with type 'a t := 'a t
    end
    module ZW (U : MONADZEROW) : sig
      include TREET with type 'a tree = 'a tree and type 'a output = 'a tree and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := U.log and type umsg := voidmsg
      include ZERO with type 'a t := 'a t
    end
    module ZE (U : MONADZEROE) : sig
      include TREET with type 'a tree = 'a tree and type 'a output = 'a tree and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := voidlog and type umsg := U.msg
      include ZERO with type 'a t := 'a t
    end
    module ZX (U : MONADZEROX) : sig
      include TREET with type 'a tree = 'a tree and type 'a output = 'a tree and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := U.log and type umsg := U.msg
      include ZERO with type 'a t := 'a t
    end
  end = struct
    module TreeBase = struct
      type 'a tree = Leaf of 'a | Branch of 'a tree * 'a tree
      let branch x y = Branch(x,y)
      let leaf x = Leaf x
      let traverse ((++) : 'b -> 'b -> 'b) (k : 'a -> 'b) (xt : 'a tree) : 'b =
        let rec aux = function
        | Leaf x -> k x
        | Branch(l, r) -> (* recursive application of k may delete a branch? *) aux l ++ aux r in
        aux xt
    end
    include TreeBase
    let map (f : 'a -> 'b) (xt : 'a tree) =
      let rec aux = function
      | Leaf x -> Leaf (f x)
      | Branch(l, r) -> Branch(aux l, aux r) in
      aux xt
    include Make.MonadFromBind(struct
      type 'a t = 'a tree
      type 'a result = 'a t let run xx = xx
      let map = `Custom map let map2 = `Generate let mapply = `Generate
      let mid = leaf
      let (>>=) xx k = traverse branch k xx
    end)
    let (++) xx yy = Branch(xx, yy)
    module Trans (U : MONADX) = struct
      include TreeBase
      type ulog = U.log
      type umsg = U.msg
      let hoist uu = U.(uu >>= fun u -> mid (Leaf u))
      let distribute k xt = traverse (U.map2 branch) (fun x -> hoist (k x)) xt
      type 'a t = 'a tree U.t
      type 'a output = 'a tree
      type 'a result = 'a tree U.result let run xx = U.run xx
      let join xtt = traverse branch ident xtt
      let (>>=) xx k = U.(>>=) xx (fun xt -> U.(>>=) (distribute k xt) (fun xtt -> U.mid (join xtt)))
      let (++) xx yy = U.(xx >>= fun xt -> yy >>= fun yt -> mid (Branch(xt,yt)))
      let hoist_map = `Generate
      let hoist_listen = `Custom (fun listener -> fun xx -> U.(listener xx >>= fun (xs,w) -> distribute (fun x -> mid (x,w)) xs))
      let hoist_pass = `Custom (fun passer xx -> passer U.(xx >>= fun xs -> let zs = traverse (fun lz rz fs -> let (l,fs') = lz fs in let (r,fs'') = rz fs' in (Branch (l,r), fs'')) (fun (x, f) fs -> (Leaf x, f % fs)) xs in let (xt, fs) = zs ident in mid (xt, fs)))
      let hoist_catch = `Generate
      let hoist_pick = `Generate
    end
    module TX (U : MONADX) = struct
      include TreeBase
      type 'a output = 'a tree
      module Trans = Trans(U)
      include Make.MonadFromT(struct
        module U = U
        include Trans
      end)
      type ulog = Trans.ulog
      type umsg = Trans.umsg
      let distribute = Trans.distribute
      let (++) = Trans.(++)
    end (* Tree.TX *)
    module ZX (U : MONADZEROX) = struct
      include TreeBase
      type 'a output = 'a tree
      module Trans = Trans(U)
      include Make.MonadFromTUZ(struct
        module U = U
        include Trans
      end)
      type ulog = Trans.ulog
      type umsg = Trans.umsg
      let distribute = Trans.distribute
      let (++) xx yy = if xx == mzero then yy else if yy == mzero then xx else Trans.(++) xx yy
    end (* Tree.ZX *)
    module T (U : MONAD) = struct include TX(struct include U type log = voidlog type msg = voidmsg end) end
    module TW (U : MONADW) = struct include TX(struct include U type msg = voidmsg end) end
    module TE (U : MONADE) = struct include TX(struct include U type log = voidlog end) end
    module Z (U : MONADZERO) = struct include ZX(struct include U type log = voidlog type msg = voidmsg end) end
    module ZW (U : MONADZEROW) = struct include ZX(struct include U type msg = voidmsg end) end
    module ZE (U : MONADZEROE) = struct include ZX(struct include U type log = voidlog end) end
  end (* Tree *)


  module Identity : sig
    include MONAD with type 'a result = 'a
  end = struct
    include Make.MonadFromComp(struct
      type 'a t = 'a
      type 'a result = 'a t let run xx = xx
      let map = `Custom (fun f x -> f x) let map2 = `Custom (fun f x y -> f x y) let mapply = `Custom (fun f x -> f x)
      let mid = ident
      let (>=>) j k = fun x -> k (j x)
    end)
  end


  (* must be parameterized on `struct type env = ... end` *)
  module Reader (E : sig type env end) : sig
    include READER with type env = E.env
    module T (U : MONAD) : READERT with type env = E.env and type 'a output = 'a and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := voidlog and type umsg := voidmsg
    module TW (U : MONADW) : READERT with type env = E.env and type 'a output = 'a and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := U.log and type umsg := voidmsg
    module TE (U : MONADE) : READERT with type env = E.env and type 'a output = 'a and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := voidlog and type umsg := U.msg
    module TX (U : MONADX) : READERT with type env = E.env and type 'a output = 'a and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := U.log and type umsg := U.msg
    module Z (U : MONADZERO) : sig
      include READERT with type env = E.env and type 'a output = 'a and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := voidlog and type umsg := voidmsg
      include ZERO with type 'a t := 'a t
    end
    module ZW (U : MONADZEROW) : sig
      include READERT with type env = E.env and type 'a output = 'a and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := U.log and type umsg := voidmsg
      include ZERO with type 'a t := 'a t
    end
    module ZE (U : MONADZEROE) : sig
      include READERT with type env = E.env and type 'a output = 'a and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := voidlog and type umsg := U.msg
      include ZERO with type 'a t := 'a t
    end
    module ZX (U : MONADZEROX) : sig
      include READERT with type env = E.env and type 'a output = 'a and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := U.log and type umsg := U.msg
      include ZERO with type 'a t := 'a t
    end
    module Z' (U : MONADZERO) : sig
      include READERT with type env = E.env and type 'a output = 'a and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := voidlog and type umsg := voidmsg
      include ZERO with type 'a t := 'a t
    end
    module Z'W (U : MONADZEROW) : sig
      include READERT with type env = E.env and type 'a output = 'a and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := U.log and type umsg := voidmsg
      include ZERO with type 'a t := 'a t
    end
    module Z'E (U : MONADZEROE) : sig
      include READERT with type env = E.env and type 'a output = 'a and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := voidlog and type umsg := U.msg
      include ZERO with type 'a t := 'a t
    end
    module Z'X (U : MONADZEROX) : sig
      include READERT with type env = E.env and type 'a output = 'a and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := U.log and type umsg := U.msg
      include ZERO with type 'a t := 'a t
    end
    module P (U : MONADPLUS) : sig
      include READERT with type env = E.env and type 'a output = 'a and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := voidlog and type umsg := voidmsg
      include PLUS with type 'a t := 'a t
    end
    module PW (U : MONADPLUSW) : sig
      include READERT with type env = E.env and type 'a output = 'a and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := U.log and type umsg := voidmsg
      include PLUS with type 'a t := 'a t
    end
    module PE (U : MONADPLUSE) : sig
      include READERT with type env = E.env and type 'a output = 'a and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := voidlog and type umsg := U.msg
      include PLUS with type 'a t := 'a t
    end
    module PX (U : MONADPLUSX) : sig
      include READERT with type env = E.env and type 'a output = 'a and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := U.log and type umsg := U.msg
      include PLUS with type 'a t := 'a t
    end
    module M (U : MONADZEROPLUS) : sig
      include READERT with type env = E.env and type 'a output = 'a and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := voidlog and type umsg := voidmsg
      include ZERO with type 'a t := 'a t
      include PLUS with type 'a t := 'a t
    end
    module MW (U : MONADZEROPLUSW) : sig
      include READERT with type env = E.env and type 'a output = 'a and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := U.log and type umsg := voidmsg
      include ZERO with type 'a t := 'a t
      include PLUS with type 'a t := 'a t
    end
    module ME (U : MONADZEROPLUSE) : sig
      include READERT with type env = E.env and type 'a output = 'a and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := voidlog and type umsg := U.msg
      include ZERO with type 'a t := 'a t
      include PLUS with type 'a t := 'a t
    end
    module MX (U : MONADZEROPLUSX) : sig
      include READERT with type env = E.env and type 'a output = 'a and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := U.log and type umsg := U.msg
      include ZERO with type 'a t := 'a t
      include PLUS with type 'a t := 'a t
    end
  end = struct
    type env = E.env
    include Make.MonadFromBind(struct
      type 'a t = env -> 'a
      type 'a result = 'a t let run xx = fun e -> xx e
      let map = `Generate let map2 = `Generate let mapply = `Generate
      let mid x = fun e -> x
      let (>>=) xx k = fun e -> let x = xx e in let xx' = k x in xx' e
    end)
    let ask = fun e -> e
    let asks selector = ask >>= (fun e -> mid (selector e)) (* may fail with Not_found *)
    let shift modifier xx = fun e -> xx (modifier e)
    module Trans (U : MONADX) = struct
      type 'a t = env -> 'a U.t
      type 'a output = 'a
      type 'a result = env -> 'a U.result let run xx = fun e -> U.run (xx e)
      type ulog = U.log
      type umsg = U.msg
      let hoist uu = fun e -> uu
      let (>>=) xx k = fun e -> U.(xx e >>= fun x -> k x e)
      let ask = U.mid
      let shift modifier xx = fun e -> xx (modifier e)
      let hoist_map = `Custom (fun f xx -> fun e -> f (xx e))
      let hoist_listen = `Generate
      let hoist_pass = `Generate
      let hoist_catch = `Custom (fun catcher xx h -> fun e -> catcher (xx e) (fun m -> (h m) e))
      let hoist_pick = `Generate
    end
    module TX (U : MONADX) = struct
      type env = E.env
      type 'a output = 'a
      module Trans = Trans(U)
      include Make.MonadFromT(struct
        module U = U
        include Trans
      end)
      type ulog = Trans.ulog
      type umsg = Trans.umsg
      let ask = Trans.ask
      let shift = Trans.shift
      let asks selector = ask >>= (fun e -> mid (selector e)) (* may fail with Not_found *)
    end (* Reader.TX *)
    module ZX (U : MONADZEROX) = struct
      type env = E.env
      type 'a output = 'a
      module Trans = Trans(U)
      include Make.MonadFromTUZ(struct
        module U = U
        include Trans
      end)
      type ulog = Trans.ulog
      type umsg = Trans.umsg
      let ask = Trans.ask
      let shift = Trans.shift
      let asks selector = ask >>= (fun e -> mid (selector e)) (* may fail with Not_found *)
    end (* Reader.ZX *)
    module Z'X (U : MONADZEROX) = struct
      type env = E.env
      type 'a output = 'a
      module Trans = Trans(U)
      include Make.MonadFromTUZ(struct
        module U = U
        include Trans
      end)
      type ulog = Trans.ulog
      type umsg = Trans.umsg
      let ask = Trans.ask
      let shift = Trans.shift
      let asks selector = ask >>= (fun e -> try mid (selector e) with Not_found -> mzero)
    end (* Reader.ZX *)
    module PX (U : MONADPLUSX) = struct
      include TX(U)
      let (++) xx yy = fun e -> U.(xx e ++ yy e)
    end (* Reader.PX *)
    module MX (U : MONADZEROPLUSX) = struct
      include ZX(U)
      let (++) xx yy = fun e -> U.(xx e ++ yy e)
    end (* Reader.MX *)
    module T (U : MONAD) = struct include TX(struct include U type log = voidlog type msg = voidmsg end) end
    module TW (U : MONADW) = struct include TX(struct include U type msg = voidmsg end) end
    module TE (U : MONADE) = struct include TX(struct include U type log = voidlog end) end
    module Z (U : MONADZERO) = struct include ZX(struct include U type log = voidlog type msg = voidmsg end) end
    module ZW (U : MONADZEROW) = struct include ZX(struct include U type msg = voidmsg end) end
    module ZE (U : MONADZEROE) = struct include ZX(struct include U type log = voidlog end) end
    module Z' (U : MONADZERO) = struct include Z'X(struct include U type log = voidlog type msg = voidmsg end) end
    module Z'W (U : MONADZEROW) = struct include Z'X(struct include U type msg = voidmsg end) end
    module Z'E (U : MONADZEROE) = struct include Z'X(struct include U type log = voidlog end) end
    module P (U : MONADPLUS) = struct include PX(struct include U type log = voidlog type msg = voidmsg end) end
    module PW (U : MONADPLUSW) = struct include PX(struct include U type msg = voidmsg end) end
    module PE (U : MONADPLUSE) = struct include PX(struct include U type log = voidlog end) end
    module M (U : MONADZEROPLUS) = struct include MX(struct include U type log = voidlog type msg = voidmsg end) end
    module MW (U : MONADZEROPLUSW) = struct include MX(struct include U type msg = voidmsg end) end
    module ME (U : MONADZEROPLUSE) = struct include MX(struct include U type log = voidlog end) end
  end (* Reader *)


  (* must be parameterized on `struct type store = ... end` *)
  module State (S : sig type store end) : sig
    include STATE with type store = S.store
    module T (U : MONAD) : STATET with type store = S.store and type 'a output = 'a * store and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := voidlog and type umsg := voidmsg
    module TW (U : MONADW) : STATET with type store = S.store and type 'a output = 'a * store and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := U.log and type umsg := voidmsg
    module TE (U : MONADE) : STATET with type store = S.store and type 'a output = 'a * store and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := voidlog and type umsg := U.msg
    module TX (U : MONADX) : STATET with type store = S.store and type 'a output = 'a * store and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := U.log and type umsg := U.msg
    module Z (U : MONADZERO) : sig
      include STATET with type store = S.store and type 'a output = 'a * store and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := voidlog and type umsg := voidmsg
      include ZERO with type 'a t := 'a t
    end
    module ZW (U : MONADZEROW) : sig
      include STATET with type store = S.store and type 'a output = 'a * store and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := U.log and type umsg := voidmsg
      include ZERO with type 'a t := 'a t
    end
    module ZE (U : MONADZEROE) : sig
      include STATET with type store = S.store and type 'a output = 'a * store and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := voidlog and type umsg := U.msg
      include ZERO with type 'a t := 'a t
    end
    module ZX (U : MONADZEROX) : sig
      include STATET with type store = S.store and type 'a output = 'a * store and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := U.log and type umsg := U.msg
      include ZERO with type 'a t := 'a t
    end
    module Z' (U : MONADZERO) : sig
      include STATET with type store = S.store and type 'a output = 'a * store and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := voidlog and type umsg := voidmsg
      include ZERO with type 'a t := 'a t
    end
    module Z'W (U : MONADZEROW) : sig
      include STATET with type store = S.store and type 'a output = 'a * store and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := U.log and type umsg := voidmsg
      include ZERO with type 'a t := 'a t
    end
    module Z'E (U : MONADZEROE) : sig
      include STATET with type store = S.store and type 'a output = 'a * store and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := voidlog and type umsg := U.msg
      include ZERO with type 'a t := 'a t
    end
    module Z'X (U : MONADZEROX) : sig
      include STATET with type store = S.store and type 'a output = 'a * store and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := U.log and type umsg := U.msg
      include ZERO with type 'a t := 'a t
    end
    module P (U : MONADPLUS) : sig
      include STATET with type store = S.store and type 'a output = 'a * store and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := voidlog and type umsg := voidmsg
      include PLUS with type 'a t := 'a t
    end
    module PW (U : MONADPLUSW) : sig
      include STATET with type store = S.store and type 'a output = 'a * store and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := U.log and type umsg := voidmsg
      include PLUS with type 'a t := 'a t
    end
    module PE (U : MONADPLUSE) : sig
      include STATET with type store = S.store and type 'a output = 'a * store and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := voidlog and type umsg := U.msg
      include PLUS with type 'a t := 'a t
    end
    module PX (U : MONADPLUSX) : sig
      include STATET with type store = S.store and type 'a output = 'a * store and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := U.log and type umsg := U.msg
      include PLUS with type 'a t := 'a t
    end
    module M (U : MONADZEROPLUS) : sig
      include STATET with type store = S.store and type 'a output = 'a * store and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := voidlog and type umsg := voidmsg
      include ZERO with type 'a t := 'a t
      include PLUS with type 'a t := 'a t
    end
    module MW (U : MONADZEROPLUSW) : sig
      include STATET with type store = S.store and type 'a output = 'a * store and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := U.log and type umsg := voidmsg
      include ZERO with type 'a t := 'a t
      include PLUS with type 'a t := 'a t
    end
    module ME (U : MONADZEROPLUSE) : sig
      include STATET with type store = S.store and type 'a output = 'a * store and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := voidlog and type umsg := U.msg
      include ZERO with type 'a t := 'a t
      include PLUS with type 'a t := 'a t
    end
    module MX (U : MONADZEROPLUSX) : sig
      include STATET with type store = S.store and type 'a output = 'a * store and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := U.log and type umsg := U.msg
      include ZERO with type 'a t := 'a t
      include PLUS with type 'a t := 'a t
    end
  end = struct
    type store = S.store
    include Make.MonadFromBind(struct
      type 'a t = store -> 'a * store
      type 'a result = 'a t let run xx = fun s -> xx s
      let map = `Generate let map2 = `Generate let mapply = `Generate
      let mid x = fun s -> x, s
      let (>>=) xx k = fun s -> let (x,s') = xx s in let xx' = k x in xx' s'
    end)
    let get = fun s -> s,s
    (* `gets viewer` is `map viewer get` *)
    let gets viewer = fun s -> viewer s, s (* may fail with Not_found *)
    let put s = fun _ -> (), s
    let modify modifier = fun s -> (), modifier s
    module Trans (U : MONADX) = struct
      type 'a t = store -> ('a * store) U.t
      type 'a output = 'a * store
      type 'a result = store -> ('a * store) U.result let run xx = fun s -> U.run (xx s)
      type ulog = U.log
      type umsg = U.msg
      let hoist uu = fun s -> U.(uu >>= fun u -> mid (u, s))
      let (>>=) xx k = fun s -> U.(xx s >>= fun (x,s') -> k x s')
      let get = fun s -> U.mid (s,s)
      let put s = fun _ -> U.mid ((), s)
      let modify modifier = fun s -> U.mid ((), modifier s)
      let hoist_map = `Custom (fun f xx -> fun s -> f (xx s))
      let hoist_listen = `Custom (fun listener xx -> fun s -> U.(listener (xx s) >>= fun ((x,s'),w) -> mid ((x,w),s')))
      let hoist_pass = `Custom (fun passer xx -> fun s -> passer U.(xx s >>= fun ((x,f),s') -> mid ((x,s'),f)))
      let hoist_catch = `Custom (fun catcher xx h -> fun s -> catcher (xx s) (fun m -> (h m) s))
      let hoist_pick = `Custom (fun picker xx -> fun s -> U.(picker (xx s) >>= fun ((x,s'),xx) -> let yy = fun s'' -> xx >>= fun (x,_) -> mid (x,s'') in mid ((x,yy),s')))
    end
    module TX (U : MONADX) = struct
      type store = S.store
      type 'a output = 'a * store
      module Trans = Trans(U)
      include Make.MonadFromT(struct
        module U = U
        include Trans
      end)
      type ulog = Trans.ulog
      type umsg = Trans.umsg
      let get = Trans.get
      let put = Trans.put
      let modify = Trans.modify
      let gets viewer = fun s -> U.mid (viewer s, s) (* may fail with Not_found *)
    end (* State.TX *)
    module ZX (U : MONADZEROX) = struct
      type store = S.store
      type 'a output = 'a * store
      module Trans = Trans(U)
      include Make.MonadFromTUZ(struct
        module U = U
        include Trans
      end)
      type ulog = Trans.ulog
      type umsg = Trans.umsg
      let get = Trans.get
      let put = Trans.put
      let modify = Trans.modify
      let gets viewer = fun s -> U.mid (viewer s, s) (* may fail with Not_found *)
    end (* State.ZX *)
    module Z'X (U : MONADZEROX) = struct
      type store = S.store
      type 'a output = 'a * store
      module Trans = Trans(U)
      include Make.MonadFromTUZ(struct
        module U = U
        include Trans
      end)
      type ulog = Trans.ulog
      type umsg = Trans.umsg
      let get = Trans.get
      let put = Trans.put
      let modify = Trans.modify
      let gets viewer = fun s -> try U.mid (viewer s, s) with Not_found -> mzero s
    end (* State.ZX *)
    module PX (U : MONADPLUSX) = struct
      include TX(U)
      let (++) xx yy = fun s -> U.(xx s ++ yy s)
    end (* State.P *)
    module MX (U : MONADZEROPLUSX) = struct
      include ZX(U)
      let (++) xx yy = fun s -> U.(xx s ++ yy s)
    end (* State.MX *)
    module T (U : MONAD) = struct include TX(struct include U type log = voidlog type msg = voidmsg end) end
    module TW (U : MONADW) = struct include TX(struct include U type msg = voidmsg end) end
    module TE (U : MONADE) = struct include TX(struct include U type log = voidlog end) end
    module Z (U : MONADZERO) = struct include ZX(struct include U type log = voidlog type msg = voidmsg end) end
    module ZW (U : MONADZEROW) = struct include ZX(struct include U type msg = voidmsg end) end
    module ZE (U : MONADZEROE) = struct include ZX(struct include U type log = voidlog end) end
    module Z' (U : MONADZERO) = struct include Z'X(struct include U type log = voidlog type msg = voidmsg end) end
    module Z'W (U : MONADZEROW) = struct include Z'X(struct include U type msg = voidmsg end) end
    module Z'E (U : MONADZEROE) = struct include Z'X(struct include U type log = voidlog end) end
    module P (U : MONADPLUS) = struct include PX(struct include U type log = voidlog type msg = voidmsg end) end
    module PW (U : MONADPLUSW) = struct include PX(struct include U type msg = voidmsg end) end
    module PE (U : MONADPLUSE) = struct include PX(struct include U type log = voidlog end) end
    module M (U : MONADZEROPLUS) = struct include MX(struct include U type log = voidlog type msg = voidmsg end) end
    module MW (U : MONADZEROPLUSW) = struct include MX(struct include U type msg = voidmsg end) end
    module ME (U : MONADZEROPLUSE) = struct include MX(struct include U type log = voidlog end) end
  end (* State *)


  (* State with a different interface; must be parameterized on `struct type value = ... end` *)
  module Ref (V : sig type value end) : sig
    include REF with type value = V.value
    type dict
    module T (U : MONAD) : REFT with type value = V.value and type 'a output = 'a * dict and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := voidlog and type umsg := voidmsg
    module TW (U : MONADW) : REFT with type value = V.value and type 'a output = 'a * dict and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := U.log and type umsg := voidmsg
    module TE (U : MONADE) : REFT with type value = V.value and type 'a output = 'a * dict and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := voidlog and type umsg := U.msg
    module TX (U : MONADX) : REFT with type value = V.value and type 'a output = 'a * dict and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := U.log and type umsg := U.msg
    module Z (U : MONADZERO) : sig
      include REFT with type value = V.value and type 'a output = 'a * dict and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := voidlog and type umsg := voidmsg
      include ZERO with type 'a t := 'a t
    end
    module ZW (U : MONADZEROW) : sig
      include REFT with type value = V.value and type 'a output = 'a * dict and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := U.log and type umsg := voidmsg
      include ZERO with type 'a t := 'a t
    end
    module ZE (U : MONADZEROE) : sig
      include REFT with type value = V.value and type 'a output = 'a * dict and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := voidlog and type umsg := U.msg
      include ZERO with type 'a t := 'a t
    end
    module ZX (U : MONADZEROX) : sig
      include REFT with type value = V.value and type 'a output = 'a * dict and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := U.log and type umsg := U.msg
      include ZERO with type 'a t := 'a t
    end
    module P (U : MONADPLUS) : sig
      include REFT with type value = V.value and type 'a output = 'a * dict and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := voidlog and type umsg := voidmsg
      include PLUS with type 'a t := 'a t
    end
    module PW (U : MONADPLUSW) : sig
      include REFT with type value = V.value and type 'a output = 'a * dict and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := U.log and type umsg := voidmsg
      include PLUS with type 'a t := 'a t
    end
    module PE (U : MONADPLUSE) : sig
      include REFT with type value = V.value and type 'a output = 'a * dict and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := voidlog and type umsg := U.msg
      include PLUS with type 'a t := 'a t
    end
    module PX (U : MONADPLUSX) : sig
      include REFT with type value = V.value and type 'a output = 'a * dict and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := U.log and type umsg := U.msg
      include PLUS with type 'a t := 'a t
    end
    module M (U : MONADZEROPLUS) : sig
      include REFT with type value = V.value and type 'a output = 'a * dict and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := voidlog and type umsg := voidmsg
      include ZERO with type 'a t := 'a t
      include PLUS with type 'a t := 'a t
    end
    module MW (U : MONADZEROPLUSW) : sig
      include REFT with type value = V.value and type 'a output = 'a * dict and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := U.log and type umsg := voidmsg
      include ZERO with type 'a t := 'a t
      include PLUS with type 'a t := 'a t
    end
    module ME (U : MONADZEROPLUSE) : sig
      include REFT with type value = V.value and type 'a output = 'a * dict and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := voidlog and type umsg := U.msg
      include ZERO with type 'a t := 'a t
      include PLUS with type 'a t := 'a t
    end
    module MX (U : MONADZEROPLUSX) : sig
      include REFT with type value = V.value and type 'a output = 'a * dict and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := U.log and type umsg := U.msg
      include ZERO with type 'a t := 'a t
      include PLUS with type 'a t := 'a t
    end
  end = struct
    type value = V.value
    type ref = int
    module D = Map.Make(struct type t = ref let compare = compare end)
    type dict = { next : ref; tree : value D.t }
    let empty = { next = 0; tree = D.empty }
    let alloc v d = d.next, { next = succ d.next; tree = D.add d.next v d.tree}
    let read (k : ref) d = D.find k d.tree
    let write (k : ref) v d = { next = d.next; tree = D.add k v d.tree }
    include Make.MonadFromBind(struct
      type 'a t = dict -> 'a * dict
      type 'a result = 'a let run xx = fst (xx empty)
      let map = `Generate let map2 = `Generate let mapply = `Generate
      let mid x = fun s -> x, s
      let (>>=) xx k = fun s -> let (x,s') = xx s in let xx' = k x in xx' s'
    end)
    let newref v = fun s -> alloc v s
    let getref k = fun s -> read k s, s (* shouldn't fail because k will have an abstract type? and we never GC *)
    let putref k v = fun s -> (), write k v s (* shouldn't allocate because k will have an abstract type *)
    module Trans (U : MONADX) = struct
      type 'a t = dict -> ('a * dict) U.t
      type 'a output = 'a * dict
      type 'a result = 'a U.result let run xx = let uu = U.(xx empty >>= fun (x,s) -> mid x) in U.run uu
      type ulog = U.log
      type umsg = U.msg
      let hoist uu = fun s -> U.(uu >>= fun u -> mid (u, s))
      let (>>=) xx k = fun s -> U.(xx s >>= fun (x,s') -> k x s')
      let newref v = fun s -> U.mid (alloc v s)
      let getref k = fun s -> U.mid (read k s, s)
      let putref k v = fun s -> U.mid ((), write k v s)
      let hoist_map = `Custom (fun f xx -> fun s -> f (xx s))
      let hoist_listen = `Custom (fun listener xx -> fun s -> U.(listener (xx s) >>= fun ((x,s'),w) -> mid ((x,w),s')))
      let hoist_pass = `Custom (fun passer xx -> fun s -> passer U.(xx s >>= fun ((x,f),s') -> mid ((x,s'),f)))
      let hoist_catch = `Custom (fun catcher xx h -> fun s -> catcher (xx s) (fun m -> (h m) s))
      let hoist_pick = `Custom (fun picker xx -> fun s -> U.(picker (xx s) >>= fun ((x,s'),xx) -> let yy = fun s'' -> xx >>= fun (x,_) -> mid (x,s'') in mid ((x,yy),s')))
    end
    module TX (U : MONADX) = struct
      type value = V.value
      type ref = int
      type 'a output = 'a * dict
      module Trans = Trans(U)
      include Make.MonadFromT(struct
        module U = U
        include Trans
      end)
      type ulog = Trans.ulog
      type umsg = Trans.umsg
      let newref = Trans.newref
      let getref = Trans.getref
      let putref = Trans.putref
    end (* Ref.TX *)
    module ZX (U : MONADZEROX) = struct
      type value = V.value
      type ref = int
      type 'a output = 'a * dict
      module Trans = Trans(U)
      include Make.MonadFromTUZ(struct
        module U = U
        include Trans
      end)
      type ulog = Trans.ulog
      type umsg = Trans.umsg
      let newref = Trans.newref
      let getref = Trans.getref
      let putref = Trans.putref
    end (* Ref.ZX *)
    module PX (U : MONADPLUSX) = struct
      include TX(U)
      let (++) xx yy = fun s -> U.(xx s ++ yy s)
    end (* Ref.P *)
    module MX (U : MONADZEROPLUSX) = struct
      include ZX(U)
      let (++) xx yy = fun s -> U.(xx s ++ yy s)
    end (* Ref.MX *)
    module T (U : MONAD) = struct include TX(struct include U type log = voidlog type msg = voidmsg end) end
    module TW (U : MONADW) = struct include TX(struct include U type msg = voidmsg end) end
    module TE (U : MONADE) = struct include TX(struct include U type log = voidlog end) end
    module Z (U : MONADZERO) = struct include ZX(struct include U type log = voidlog type msg = voidmsg end) end
    module ZW (U : MONADZEROW) = struct include ZX(struct include U type msg = voidmsg end) end
    module ZE (U : MONADZEROE) = struct include ZX(struct include U type log = voidlog end) end
    module P (U : MONADPLUS) = struct include PX(struct include U type log = voidlog type msg = voidmsg end) end
    module PW (U : MONADPLUSW) = struct include PX(struct include U type msg = voidmsg end) end
    module PE (U : MONADPLUSE) = struct include PX(struct include U type log = voidlog end) end
    module M (U : MONADZEROPLUS) = struct include MX(struct include U type log = voidlog type msg = voidmsg end) end
    module MW (U : MONADZEROPLUSW) = struct include MX(struct include U type msg = voidmsg end) end
    module ME (U : MONADZEROPLUSE) = struct include MX(struct include U type log = voidlog end) end
  end (* Ref *)


  (* must be parameterized on `struct type log = ... end` *)
  module Writer (W : sig type log val empty : log val append : log -> log -> log end) : sig
    include WRITER with type log = W.log
    module T (U : MONAD) : WRITERT with type log = W.log and type 'a output = 'a * W.log and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type umsg := voidmsg
    module TE (U : MONADE) : WRITERT with type log = W.log and type 'a output = 'a * W.log and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type umsg := U.msg
    module Z (U : MONADZERO) : sig
      include WRITERT with type log = W.log and type 'a output = 'a * W.log and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type umsg := voidmsg
      include ZERO with type 'a t := 'a t
    end
    module ZE (U : MONADZEROE) : sig
      include WRITERT with type log = W.log and type 'a output = 'a * W.log and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type umsg := U.msg
      include ZERO with type 'a t := 'a t
    end
    module P (U : MONADPLUS) : sig
      include WRITERT with type log = W.log and type 'a output = 'a * W.log and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type umsg := voidmsg
      include PLUS with type 'a t := 'a t
    end
    module PE (U : MONADPLUSE) : sig
      include WRITERT with type log = W.log and type 'a output = 'a * W.log and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type umsg := U.msg
      include PLUS with type 'a t := 'a t
    end
    module M (U : MONADZEROPLUS) : sig
      include WRITERT with type log = W.log and type 'a output = 'a * W.log and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type umsg := voidmsg
      include ZERO with type 'a t := 'a t
      include PLUS with type 'a t := 'a t
    end
    module ME (U : MONADZEROPLUSE) : sig
      include WRITERT with type log = W.log and type 'a output = 'a * W.log and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type umsg := U.msg
      include ZERO with type 'a t := 'a t
      include PLUS with type 'a t := 'a t
    end
  end = struct
    type log = W.log
    include Make.MonadFromBind(struct
      type 'a t = 'a * log
      type 'a result = 'a t let run xx = xx
      let map = `Generate let map2 = `Generate let mapply = `Generate
      let mid x = x, W.empty
      let (>>=) (x,w) k = let (y,w') = k x in (y, W.append w w')
    end)
    let listen (x,w) = (x,w), w
    let listens selector xx = listen xx >>= fun (x,w) -> mid (x,selector w) (* filter listen through selector *)
    let tell entries = (), entries (* add to log *)
    let pass ((x,c),w) = (x, c w) (* usually use censor *)
    let censor c xx = pass (xx >>= fun x -> mid (x,c)) (* ==> (x, c w) *)
    module Trans (U : MONADE) = struct
      type 'a t = ('a * log) U.t
      type 'a output = 'a * log
      type 'a result = ('a * log) U.result let run xx = U.run xx
      type umsg = U.msg
      let hoist uu = U.(uu >>= fun u -> mid (u, W.empty))
      let (>>=) xx k = U.(xx >>= fun (x,w) -> k x >>= fun (y,w') -> mid (y, W.append w w'))
      let listen xx = U.(xx >>= fun (x,w) -> mid ((x,w),w))
      let tell entries = U.mid ((), entries)
      let pass xx = U.(xx >>= fun ((x,c),w) -> mid (x, c w))
      let hoist_map = `Generate
      let hoist_listen = `Generate
      let hoist_pass = `Generate
      let hoist_catch = `Generate
      let hoist_pick = `Generate
    end
    module TE (U : MONADE) = struct
      type log = W.log
      type 'a output = 'a * log
      module Trans = Trans(U)
      include Make.MonadFromT(struct
        module U = U
        type ulog = log
        include Trans
      end)
      type umsg = Trans.umsg
      let listen = Trans.listen
      let tell = Trans.tell
      let listens selector xx = listen xx >>= fun (x,w) -> mid (x,selector w)
      let pass = Trans.pass
      let censor c xx = Trans.pass (xx >>= fun x -> mid (x,c))
    end (* Writer.TE *)
    module ZE (U : MONADZEROE) = struct
      type log = W.log
      type 'a output = 'a * log
      module Trans = Trans(U)
      include Make.MonadFromTUZ(struct
        module U = U
        type ulog = log
        include Trans
      end)
      type umsg = Trans.umsg
      let listen = Trans.listen
      let tell = Trans.tell
      let listens selector xx = listen xx >>= fun (x,w) -> mid (x,selector w)
      let pass = Trans.pass
      let censor c xx = Trans.pass (xx >>= fun x -> mid (x,c))
    end (* Writer.ZE *)
    module PE (U : MONADPLUSE) = struct
      include TE(U)
      let (++) xx yy = U.(xx ++ yy)
    end (* Writer.PE *)
    module ME (U : MONADZEROPLUSE) = struct
      include ZE(U)
      let (++) xx yy = U.(xx ++ yy)
    end (* Writer.ME *)
    module T (U : MONAD) = struct include TE(struct include U type msg = voidmsg end) end
    module Z (U : MONADZERO) = struct include ZE(struct include U type msg = voidmsg end) end
    module P (U : MONADPLUS) = struct include PE(struct include U type msg = voidmsg end) end
    module M (U : MONADZEROPLUS) = struct include ME(struct include U type msg = voidmsg end) end
  end (* Writer *)


  (* must be parameterized on `struct type msg = ... end` *)
  module Error (E : sig type msg exception Exn of msg (* Exn used only by T *) end) : sig
    type msg = E.msg
    type 'a error = Error of msg | OK of 'a
    include ERROR with type msg := msg and type 'a error := 'a error
    module T (U : MONAD) : ERRORT with type msg = E.msg and type 'a error = 'a error and type 'a output = 'a error and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := voidlog
    module TW (U : MONADW) : ERRORT with type msg = E.msg and type 'a error = 'a error and type 'a output = 'a error and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := U.log
    module Z (U : MONADZERO) : sig
      include ERRORT with type msg = E.msg and type 'a error = 'a error and type 'a output = 'a error and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := voidlog
      include ZERO with type 'a t := 'a t
    end
    module ZW (U : MONADZEROW) : sig
      include ERRORT with type msg = E.msg and type 'a error = 'a error and type 'a output = 'a error and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := U.log
      include ZERO with type 'a t := 'a t
    end
    module P (U : MONADPLUS) : sig
      include ERRORT with type msg = E.msg and type 'a error = 'a error and type 'a output = 'a error and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := voidlog
      include PLUS with type 'a t := 'a t
    end
    module PW (U : MONADPLUSW) : sig
      include ERRORT with type msg = E.msg and type 'a error = 'a error and type 'a output = 'a error and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := U.log
      include PLUS with type 'a t := 'a t
    end
    module M (U : MONADZEROPLUS) : sig
      include ERRORT with type msg = E.msg and type 'a error = 'a error and type 'a output = 'a error and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := voidlog
      include ZERO with type 'a t := 'a t
      include PLUS with type 'a t := 'a t
    end
    module MW (U : MONADZEROPLUSW) : sig
      include ERRORT with type msg = E.msg and type 'a error = 'a error and type 'a output = 'a error and type 'a uresult := 'a U.result and type 'a ut := 'a U.t and type ulog := U.log
      include ZERO with type 'a t := 'a t
      include PLUS with type 'a t := 'a t
    end
  end = struct
    module ErrorBase = struct
      type msg = E.msg
      type 'a error = Error of msg | OK of 'a
    end
    include ErrorBase
    include Make.MonadFromBind(struct
      type 'a t = 'a error
      type 'a result = 'a t
      (* note that M.run doesn't raise *)
      let run xx = xx
      let map = `Generate let map2 = `Generate let mapply = `Generate
      let mid x = OK x
      let (>>=) xx k = match xx with OK x -> k x | Error e -> Error e
    end)
    let throw e = Error e
    let catch xx handler = match xx with OK _ -> xx | Error e -> handler e
    module Trans (U : MONADW) = struct
      include ErrorBase
      type 'a t = 'a error U.t
      type 'a output = 'a error
      type 'a result = 'a U.result
      type ulog = U.log
      let hoist uu = U.(uu >>= fun u -> mid (OK u))
      let (>>=) xx k = U.(xx >>= function OK x -> k x | Error e -> mid (Error e))
      let throw e = U.mid (Error e)
      let catch xx handler = U.(xx >>= function OK _ as x -> mid x | Error e -> handler e)
      let hoist_map = `Generate
      let hoist_listen = `Custom (fun listener -> fun xx -> U.(listener xx >>= function (Error e, _) -> mid (Error e) | (OK y, w) -> mid (OK (y, w))))
      let hoist_pass = `Custom (fun passer -> fun xx -> passer U.(xx >>= function Error e -> mid (Error e, ident) | OK (x, f) -> mid (OK x, f)))
      let hoist_catch = `Generate
      let hoist_pick = `Custom (fun picker -> fun xx -> U.(picker xx >>= function (Error e, _) -> mid (Error e) | (OK x, yy) -> mid (OK (x, yy))))
    end
    module TW (U : MONADW) = struct
      include ErrorBase
      type 'a output = 'a error
      module Trans = Trans(U)
      include Make.MonadFromT(struct
        module U = U
        type umsg = msg
        include Trans
        (* note that T.run does raise *)
        let run xx = let uu = U.(xx >>= function OK x -> mid x | Error e -> raise (E.Exn e)) in U.run uu
      end)
      type ulog = Trans.ulog
      let throw = Trans.throw
      let catch = Trans.catch
    end (* Error.TW *)
    module ZW (U : MONADZEROW) = struct
      include ErrorBase
      type 'a output = 'a error
      module Trans = Trans(U)
      include Make.MonadFromTUZ(struct
        module U = U
        type umsg = msg
        include Trans
        (* we recover from error by using U's mzero; but this discards the error msg *)
        let run xx = let uu = U.(xx >>= function OK x -> mid x | Error e -> mzero) in U.run uu
      end)
      type ulog = Trans.ulog
      let throw = Trans.throw
      let catch = Trans.catch
    end (* Error.ZW *)
    module PW (U : MONADPLUSW) = struct
      include TW(U)
      let (++) xx yy = U.(xx ++ yy)
    end (* Error.PW *)
    module MW (U : MONADZEROPLUSW) = struct
      include ZW(U)
      let (++) xx yy = U.(xx ++ yy)
    end (* Error.MW *)
    module T (U : MONAD) = struct include TW(struct include U type log = voidlog end) end
    module Z (U : MONADZERO) = struct include ZW(struct include U type log = voidlog end) end
    module P (U : MONADPLUS) = struct include PW(struct include U type log = voidlog end) end
    module M (U : MONADZEROPLUS) = struct include MW(struct include U type log = voidlog end) end
  end (* Error *)


  (* predefine some common instances *)

  module Writer1 = Writer(struct
    type log = string
    let empty = ""
    let append s1 s2 = if s2 = "" then s1 else if s1 = "" then s2 else s1 ^ "\n" ^ s2
  end)

  module Writer2 = struct
    include Writer(struct
      type log = string list
      let empty = []
      let append s1 s2 = Juli8.List.append s2 s1
    end)
    let tell_string s = tell [s]
    let tell entries = tell (Juli8.List.rev entries)
    let run xx = let (x,w) = run xx in (x, Juli8.List.rev w)
  end

  module Failure = Error(struct
    type msg = string
    exception Exn = Failure
  end)


  (*
    Here is how to wrap a ListT around an underlying Monad that has the Reader interface:
      module L = Monad.List
      module R = Monad.Reader(struct type env = char end)
      module LR = struct include L.T(R) include Monad.Derive.Reader(R)(L.T(R)) end
  *)
  module Derive = struct

    module Option (U : OPTIONBASE) (T : MONADT with type 'a ut := 'a U.t and type ulog := voidlog and type umsg := voidmsg) : OPTIONBASE with type 'a t := 'a T.t = struct end
    module OptionW (U : sig include OPTIONBASE type log end) (T : MONADT with type 'a ut := 'a U.t and type ulog := U.log and type umsg := voidmsg) : OPTIONBASE with type 'a t := 'a T.t = struct end
    module Option_toW (U : OPTIONBASE) (T : MONADTW with type 'a ut := 'a U.t and type umsg := voidmsg) : OPTIONBASE with type 'a t := 'a T.t = struct end
    module OptionE (U : sig include OPTIONBASE type msg end) (T : MONADT with type 'a ut := 'a U.t and type ulog := voidlog and type umsg := U.msg) : OPTIONBASE with type 'a t := 'a T.t = struct end
    module Option_toE (U : OPTIONBASE) (T : MONADTE with type 'a ut := 'a U.t and type ulog := voidlog) : OPTIONBASE with type 'a t := 'a T.t = struct end
    module OptionW_toE (U : sig include OPTIONBASE type log end) (T : MONADTE with type 'a ut := 'a U.t and type ulog := U.log) : OPTIONBASE with type 'a t := 'a T.t = struct end
    module OptionE_toW (U : sig include OPTIONBASE type msg end) (T : MONADTW with type 'a ut := 'a U.t and type umsg := U.msg) : OPTIONBASE with type 'a t := 'a T.t = struct end
    module OptionX (U : sig include OPTIONBASE type log type msg end) (T : MONADT with type 'a ut := 'a U.t and type ulog := U.log and type umsg := U.msg) : OPTIONBASE with type 'a t := 'a T.t = struct end

    module Option_dbl (U : OPTIONBASE) (T : OPTION_ with type 'a ut := 'a U.t and type ulog := voidlog and type umsg := voidmsg) = struct end
    module OptionW_dbl (U : sig include OPTIONBASE type log end) (T : OPTION_ with type 'a ut := 'a U.t and type ulog := U.log and type umsg := voidmsg) = struct end
    module OptionE_dbl (U : sig include OPTIONBASE type msg end) (T : OPTION_ with type 'a ut := 'a U.t and type ulog := voidlog and type umsg := U.msg) = struct end
    module OptionX_dbl (U : sig include OPTIONBASE type log type msg end) (T : OPTION_ with type 'a ut := 'a U.t and type ulog := U.log and type umsg := U.msg) = struct end


    module List (U : LISTBASE) (T : MONADT with type 'a ut := 'a U.t and type ulog := voidlog and type umsg := voidmsg) : LISTBASE with type 'a t := 'a T.t = struct
      let pick xx = T.hoist_pick (U.pick) xx
    end
    module ListW (U : sig include LISTBASE type log end) (T : MONADT with type 'a ut := 'a U.t and type ulog := U.log and type umsg := voidmsg) : LISTBASE with type 'a t := 'a T.t = struct
      let pick xx = T.hoist_pick (U.pick) xx
    end
    module List_toW (U : LISTBASE) (T : MONADTW with type 'a ut := 'a U.t and type umsg := voidmsg) : LISTBASE with type 'a t := 'a T.t = struct
      let pick xx = T.hoist_pick (U.pick) xx
    end
    module ListE (U : sig include LISTBASE type msg end) (T : MONADT with type 'a ut := 'a U.t and type ulog := voidlog and type umsg := U.msg) : LISTBASE with type 'a t := 'a T.t = struct
      let pick xx = T.hoist_pick (U.pick) xx
    end
    module List_toE (U : LISTBASE) (T : MONADTE with type 'a ut := 'a U.t and type ulog := voidlog) : LISTBASE with type 'a t := 'a T.t = struct
      let pick xx = T.hoist_pick (U.pick) xx
    end
    module ListW_toE (U : sig include LISTBASE type log end) (T : MONADTE with type 'a ut := 'a U.t and type ulog := U.log) : LISTBASE with type 'a t := 'a T.t = struct
      let pick xx = T.hoist_pick (U.pick) xx
    end
    module ListE_toW (U : sig include LISTBASE type msg end) (T : MONADTW with type 'a ut := 'a U.t and type umsg := U.msg) : LISTBASE with type 'a t := 'a T.t = struct
      let pick xx = T.hoist_pick (U.pick) xx
    end
    module ListX (U : sig include LISTBASE type log type msg end) (T : MONADT with type 'a ut := 'a U.t and type ulog := U.log and type umsg := U.msg) : LISTBASE with type 'a t := 'a T.t = struct
      let pick xx = T.hoist_pick (U.pick) xx
    end

    module List_dbl (U : LISTBASE) (T : LIST_ with type 'a ut := 'a U.t and type ulog := voidlog and type umsg := voidmsg) = struct
      let upick xx = T.hoist_pick (U.pick) xx
    end
    module ListW_dbl (U : sig include LISTBASE type log end) (T : LIST_ with type 'a ut := 'a U.t and type ulog := U.log and type umsg := voidmsg) = struct
      let upick xx = T.hoist_pick (U.pick) xx
    end
    module ListE_dbl (U : sig include LISTBASE type msg end) (T : LIST_ with type 'a ut := 'a U.t and type ulog := voidlog and type umsg := U.msg) = struct
      let upick xx = T.hoist_pick (U.pick) xx
    end
    module ListX_dbl (U : sig include LISTBASE type log type msg end) (T : LIST_ with type 'a ut := 'a U.t and type ulog := U.log and type umsg := U.msg) = struct
      let upick xx = T.hoist_pick (U.pick) xx
    end

    module Tree (U : TREEBASE) (T : MONADT with type 'a ut := 'a U.t and type ulog := voidlog and type umsg := voidmsg) : TREEBASE with type 'a t := 'a T.t = struct
      type 'a tree = 'a U.tree
      let branch = U.branch
      let leaf = U.leaf
      let traverse = U.traverse
    end
    module TreeW (U : sig include TREEBASE type log end) (T : MONADT with type 'a ut := 'a U.t and type ulog := U.log and type umsg := voidmsg) : TREEBASE with type 'a t := 'a T.t = struct
      type 'a tree = 'a U.tree
      let branch = U.branch
      let leaf = U.leaf
      let traverse = U.traverse
    end
    module Tree_toW (U : TREEBASE) (T : MONADTW with type 'a ut := 'a U.t and type umsg := voidmsg) : TREEBASE with type 'a t := 'a T.t = struct
      type 'a tree = 'a U.tree
      let branch = U.branch
      let leaf = U.leaf
      let traverse = U.traverse
    end
    module TreeE (U : sig include TREEBASE type msg end) (T : MONADT with type 'a ut := 'a U.t and type ulog := voidlog and type umsg := U.msg) : TREEBASE with type 'a t := 'a T.t = struct
      type 'a tree = 'a U.tree
      let branch = U.branch
      let leaf = U.leaf
      let traverse = U.traverse
    end
    module Tree_toE (U : TREEBASE) (T : MONADTE with type 'a ut := 'a U.t and type ulog := voidlog) : TREEBASE with type 'a t := 'a T.t = struct
      type 'a tree = 'a U.tree
      let branch = U.branch
      let leaf = U.leaf
      let traverse = U.traverse
    end
    module TreeW_toE (U : sig include TREEBASE type log end) (T : MONADTE with type 'a ut := 'a U.t and type ulog := U.log) : TREEBASE with type 'a t := 'a T.t = struct
      type 'a tree = 'a U.tree
      let branch = U.branch
      let leaf = U.leaf
      let traverse = U.traverse
    end
    module TreeE_toW (U : sig include TREEBASE type msg end) (T : MONADTW with type 'a ut := 'a U.t and type umsg := U.msg) : TREEBASE with type 'a t := 'a T.t = struct
      type 'a tree = 'a U.tree
      let branch = U.branch
      let leaf = U.leaf
      let traverse = U.traverse
    end
    module TreeX (U : sig include TREEBASE type log type msg end) (T : MONADT with type 'a ut := 'a U.t and type ulog := U.log and type umsg := U.msg) : TREEBASE with type 'a t := 'a T.t = struct
      type 'a tree = 'a U.tree
      let branch = U.branch
      let leaf = U.leaf
      let traverse = U.traverse
    end

    module Tree_dbl (U : TREEBASE) (T : TREE_ with type 'a ut := 'a U.t and type ulog := voidlog and type umsg := voidmsg) = struct
      type 'a utree = 'a U.tree
      let ubranch = U.branch
      let uleaf = U.leaf
      let utraverse = U.traverse
    end
    module TreeW_dbl (U : sig include TREEBASE type log end) (T : TREE_ with type 'a ut := 'a U.t and type ulog := U.log and type umsg := voidmsg) = struct
      type 'a utree = 'a U.tree
      let ubranch = U.branch
      let uleaf = U.leaf
      let utraverse = U.traverse
    end
    module TreeE_dbl (U : sig include TREEBASE type msg end) (T : TREE_ with type 'a ut := 'a U.t and type ulog := voidlog and type umsg := U.msg) = struct
      type 'a utree = 'a U.tree
      let ubranch = U.branch
      let uleaf = U.leaf
      let utraverse = U.traverse
    end
    module TreeX_dbl (U : sig include TREEBASE type log type msg end) (T : TREET with type 'a ut := 'a U.t and type ulog := U.log and type umsg := U.msg) = struct
      type 'a utree = 'a U.tree
      let ubranch = U.branch
      let uleaf = U.leaf
      let utraverse = U.traverse
    end

    module Reader (U : READERBASE) (T : MONADT with type 'a ut := 'a U.t and type ulog := voidlog and type umsg := voidmsg) : READERBASE with type env = U.env and type 'a t := 'a T.t = struct
      type env = U.env
      let ask = T.hoist U.ask
      let asks selector = T.hoist (U.asks selector)
      let shift modifier xx = T.hoist_map (U.shift modifier) xx
    end
    module ReaderW (U : sig include READERBASE type log end) (T : MONADT with type 'a ut := 'a U.t and type ulog := U.log and type umsg := voidmsg) : READERBASE with type env = U.env and type 'a t := 'a T.t = struct
      type env = U.env
      let ask = T.hoist U.ask
      let asks selector = T.hoist (U.asks selector)
      let shift modifier xx = T.hoist_map (U.shift modifier) xx
    end
    module Reader_toW (U : READERBASE) (T : MONADTW with type 'a ut := 'a U.t and type umsg := voidmsg) : READERBASE with type env = U.env and type 'a t := 'a T.t = struct
      type env = U.env
      let ask = T.hoist U.ask
      let asks selector = T.hoist (U.asks selector)
      let shift modifier xx = T.hoist_map (U.shift modifier) xx
    end
    module ReaderE (U : sig include READERBASE type msg end) (T : MONADT with type 'a ut := 'a U.t and type ulog := voidlog and type umsg := U.msg) : READERBASE with type env = U.env and type 'a t := 'a T.t = struct
      type env = U.env
      let ask = T.hoist U.ask
      let asks selector = T.hoist (U.asks selector)
      let shift modifier xx = T.hoist_map (U.shift modifier) xx
    end
    module Reader_toE (U : READERBASE) (T : MONADTE with type 'a ut := 'a U.t and type ulog := voidlog) : READERBASE with type env = U.env and type 'a t := 'a T.t = struct
      type env = U.env
      let ask = T.hoist U.ask
      let asks selector = T.hoist (U.asks selector)
      let shift modifier xx = T.hoist_map (U.shift modifier) xx
    end
    module ReaderW_toE (U : sig include READERBASE type log end) (T : MONADTE with type 'a ut := 'a U.t and type ulog := U.log) : READERBASE with type env = U.env and type 'a t := 'a T.t = struct
      type env = U.env
      let ask = T.hoist U.ask
      let asks selector = T.hoist (U.asks selector)
      let shift modifier xx = T.hoist_map (U.shift modifier) xx
    end
    module ReaderE_toW (U : sig include READERBASE type msg end) (T : MONADTW with type 'a ut := 'a U.t and type umsg := U.msg) : READERBASE with type env = U.env and type 'a t := 'a T.t = struct
      type env = U.env
      let ask = T.hoist U.ask
      let asks selector = T.hoist (U.asks selector)
      let shift modifier xx = T.hoist_map (U.shift modifier) xx
    end
    module ReaderX (U : sig include READERBASE type log type msg end) (T : MONADT with type 'a ut := 'a U.t and type ulog := U.log and type umsg := U.msg) : READERBASE with type env = U.env and type 'a t := 'a T.t = struct
      type env = U.env
      let ask = T.hoist U.ask
      let asks selector = T.hoist (U.asks selector)
      let shift modifier xx = T.hoist_map (U.shift modifier) xx
    end

    module Reader_dbl (U : READERBASE) (T : READER_ with type 'a ut := 'a U.t and type ulog := voidlog and type umsg := voidmsg) = struct
      type uenv = U.env
      let uask = T.hoist U.ask
      let uasks selector = T.hoist (U.asks selector)
      let ushift modifier xx = T.hoist_map (U.shift modifier) xx
    end
    module ReaderW_dbl (U : sig include READERBASE type log end) (T : READER_ with type 'a ut := 'a U.t and type ulog := U.log and type umsg := voidmsg) = struct
      type uenv = U.env
      let uask = T.hoist U.ask
      let uasks selector = T.hoist (U.asks selector)
      let ushift modifier xx = T.hoist_map (U.shift modifier) xx
    end
    module ReaderE_dbl (U : sig include READERBASE type msg end) (T : READER_ with type 'a ut := 'a U.t and type ulog := voidlog and type umsg := U.msg) = struct
      type uenv = U.env
      let uask = T.hoist U.ask
      let uasks selector = T.hoist (U.asks selector)
      let ushift modifier xx = T.hoist_map (U.shift modifier) xx
    end
    module ReaderX_dbl (U : sig include READERBASE type log type msg end) (T : READER_ with type 'a ut := 'a U.t and type ulog := U.log and type umsg := U.msg) = struct
      type uenv = U.env
      let uask = T.hoist U.ask
      let uasks selector = T.hoist (U.asks selector)
      let ushift modifier xx = T.hoist_map (U.shift modifier) xx
    end

    module Reader2 (U : READER2BASE) (T : MONADT with type 'a ut := 'a U.t and type ulog := voidlog and type umsg := voidmsg) : READER2BASE with type env = U.env and type uenv = U.uenv and type 'a t := 'a T.t = struct
      type env = U.env
      let ask = T.hoist U.ask
      let asks selector = T.hoist (U.asks selector)
      let shift modifier xx = T.hoist_map (U.shift modifier) xx
      type uenv = U.uenv
      let uask = T.hoist U.uask
      let uasks selector = T.hoist (U.uasks selector)
      let ushift modifier xx = T.hoist_map (U.ushift modifier) xx
    end
    module Reader2W (U : sig include READER2BASE type log end) (T : MONADT with type 'a ut := 'a U.t and type ulog := U.log and type umsg := voidmsg) : READER2BASE with type env = U.env and type uenv = U.uenv and type 'a t := 'a T.t = struct
      type env = U.env
      let ask = T.hoist U.ask
      let asks selector = T.hoist (U.asks selector)
      let shift modifier xx = T.hoist_map (U.shift modifier) xx
      type uenv = U.uenv
      let uask = T.hoist U.uask
      let uasks selector = T.hoist (U.uasks selector)
      let ushift modifier xx = T.hoist_map (U.ushift modifier) xx
    end
    module Reader2_toW (U : READER2BASE) (T : MONADTW with type 'a ut := 'a U.t and type umsg := voidmsg) : READER2BASE with type env = U.env and type uenv = U.uenv and type 'a t := 'a T.t = struct
      type env = U.env
      let ask = T.hoist U.ask
      let asks selector = T.hoist (U.asks selector)
      let shift modifier xx = T.hoist_map (U.shift modifier) xx
      type uenv = U.uenv
      let uask = T.hoist U.uask
      let uasks selector = T.hoist (U.uasks selector)
      let ushift modifier xx = T.hoist_map (U.ushift modifier) xx
    end
    module Reader2E (U : sig include READER2BASE type msg end) (T : MONADT with type 'a ut := 'a U.t and type ulog := voidlog and type umsg := U.msg) : READER2BASE with type env = U.env and type 'a t := 'a T.t = struct
      type env = U.env
      let ask = T.hoist U.ask
      let asks selector = T.hoist (U.asks selector)
      let shift modifier xx = T.hoist_map (U.shift modifier) xx
      type uenv = U.uenv
      let uask = T.hoist U.uask
      let uasks selector = T.hoist (U.uasks selector)
      let ushift modifier xx = T.hoist_map (U.ushift modifier) xx
    end
    module Reader2_toE (U : READER2BASE) (T : MONADTE with type 'a ut := 'a U.t and type ulog := voidlog) : READER2BASE with type env = U.env and type uenv = U.uenv and type 'a t := 'a T.t = struct
      type env = U.env
      let ask = T.hoist U.ask
      let asks selector = T.hoist (U.asks selector)
      let shift modifier xx = T.hoist_map (U.shift modifier) xx
      type uenv = U.uenv
      let uask = T.hoist U.uask
      let uasks selector = T.hoist (U.uasks selector)
      let ushift modifier xx = T.hoist_map (U.ushift modifier) xx
    end
    module Reader2W_toE (U : sig include READER2BASE type log end) (T : MONADTE with type 'a ut := 'a U.t and type ulog := U.log) : READER2BASE with type env = U.env and type uenv = U.uenv and type 'a t := 'a T.t = struct
      type env = U.env
      let ask = T.hoist U.ask
      let asks selector = T.hoist (U.asks selector)
      let shift modifier xx = T.hoist_map (U.shift modifier) xx
      type uenv = U.uenv
      let uask = T.hoist U.uask
      let uasks selector = T.hoist (U.uasks selector)
      let ushift modifier xx = T.hoist_map (U.ushift modifier) xx
    end
    module Reader2E_toW (U : sig include READER2BASE type msg end) (T : MONADTW with type 'a ut := 'a U.t and type umsg := U.msg) : READER2BASE with type env = U.env and type uenv = U.uenv and type 'a t := 'a T.t = struct
      type env = U.env
      let ask = T.hoist U.ask
      let asks selector = T.hoist (U.asks selector)
      let shift modifier xx = T.hoist_map (U.shift modifier) xx
      type uenv = U.uenv
      let uask = T.hoist U.uask
      let uasks selector = T.hoist (U.uasks selector)
      let ushift modifier xx = T.hoist_map (U.ushift modifier) xx
    end
    module Reader2X (U : sig include READER2BASE type log type msg end) (T : MONADT with type 'a ut := 'a U.t and type ulog := U.log and type umsg := U.msg) : READER2BASE with type env = U.env and type uenv = U.uenv and type 'a t := 'a T.t = struct
      type env = U.env
      let ask = T.hoist U.ask
      let asks selector = T.hoist (U.asks selector)
      let shift modifier xx = T.hoist_map (U.shift modifier) xx
      type uenv = U.uenv
      let uask = T.hoist U.uask
      let uasks selector = T.hoist (U.uasks selector)
      let ushift modifier xx = T.hoist_map (U.ushift modifier) xx
    end

    module State (U : STATEBASE) (T : MONADT with type 'a ut := 'a U.t and type ulog := voidlog and type umsg := voidmsg) : STATEBASE with type store = U.store and type 'a t := 'a T.t = struct
      type store = U.store
      let get = T.hoist U.get
      let gets viewer = T.hoist (U.gets viewer)
      let put s = T.hoist (U.put s)
      let modify modifier = T.hoist (U.modify modifier)
    end
    module StateW (U : sig include STATEBASE type log end) (T : MONADT with type 'a ut := 'a U.t and type ulog := U.log and type umsg := voidmsg) : STATEBASE with type store = U.store and type 'a t := 'a T.t = struct
      type store = U.store
      let get = T.hoist U.get
      let gets viewer = T.hoist (U.gets viewer)
      let put s = T.hoist (U.put s)
      let modify modifier = T.hoist (U.modify modifier)
    end
    module State_toW (U : STATEBASE) (T : MONADTW with type 'a ut := 'a U.t and type umsg := voidmsg) : STATEBASE with type store = U.store and type 'a t := 'a T.t = struct
      type store = U.store
      let get = T.hoist U.get
      let gets viewer = T.hoist (U.gets viewer)
      let put s = T.hoist (U.put s)
      let modify modifier = T.hoist (U.modify modifier)
    end
    module StateE (U : sig include STATEBASE type msg end) (T : MONADT with type 'a ut := 'a U.t and type ulog := voidlog and type umsg := U.msg) : STATEBASE with type store = U.store and type 'a t := 'a T.t = struct
      type store = U.store
      let get = T.hoist U.get
      let gets viewer = T.hoist (U.gets viewer)
      let put s = T.hoist (U.put s)
      let modify modifier = T.hoist (U.modify modifier)
    end
    module State_toE (U : STATEBASE) (T : MONADTE with type 'a ut := 'a U.t and type ulog := voidlog) : STATEBASE with type store = U.store and type 'a t := 'a T.t = struct
      type store = U.store
      let get = T.hoist U.get
      let gets viewer = T.hoist (U.gets viewer)
      let put s = T.hoist (U.put s)
      let modify modifier = T.hoist (U.modify modifier)
    end
    module StateW_toE (U : sig include STATEBASE type log end) (T : MONADTE with type 'a ut := 'a U.t and type ulog := U.log) : STATEBASE with type store = U.store and type 'a t := 'a T.t = struct
      type store = U.store
      let get = T.hoist U.get
      let gets viewer = T.hoist (U.gets viewer)
      let put s = T.hoist (U.put s)
      let modify modifier = T.hoist (U.modify modifier)
    end
    module StateE_toW (U : sig include STATEBASE type msg end) (T : MONADTW with type 'a ut := 'a U.t and type umsg := U.msg) : STATEBASE with type store = U.store and type 'a t := 'a T.t = struct
      type store = U.store
      let get = T.hoist U.get
      let gets viewer = T.hoist (U.gets viewer)
      let put s = T.hoist (U.put s)
      let modify modifier = T.hoist (U.modify modifier)
    end
    module StateX (U : sig include STATEBASE type log type msg end) (T : MONADT with type 'a ut := 'a U.t and type ulog := U.log and type umsg := U.msg) : STATEBASE with type store = U.store and type 'a t := 'a T.t = struct
      type store = U.store
      let get = T.hoist U.get
      let gets viewer = T.hoist (U.gets viewer)
      let put s = T.hoist (U.put s)
      let modify modifier = T.hoist (U.modify modifier)
    end

    module State_dbl (U : STATEBASE) (T : STATE_ with type 'a ut := 'a U.t and type ulog := voidlog and type umsg := voidmsg) = struct
      type ustore = U.store
      let uget = T.hoist U.get
      let ugets viewer = T.hoist (U.gets viewer)
      let uput s = T.hoist (U.put s)
      let umodify modifier = T.hoist (U.modify modifier)
    end
    module StateW_dbl (U : sig include STATEBASE type log end) (T : STATE_ with type 'a ut := 'a U.t and type ulog := U.log and type umsg := voidmsg) = struct
      type ustore = U.store
      let uget = T.hoist U.get
      let ugets viewer = T.hoist (U.gets viewer)
      let uput s = T.hoist (U.put s)
      let umodify modifier = T.hoist (U.modify modifier)
    end
    module StateE_dbl (U : sig include STATEBASE type msg end) (T : STATE_ with type 'a ut := 'a U.t and type ulog := voidlog and type umsg := U.msg) = struct
      type ustore = U.store
      let uget = T.hoist U.get
      let ugets viewer = T.hoist (U.gets viewer)
      let uput s = T.hoist (U.put s)
      let umodify modifier = T.hoist (U.modify modifier)
    end
    module StateX_dbl (U : sig include STATEBASE type log type msg end) (T : STATE_ with type 'a ut := 'a U.t and type ulog := U.log and type umsg := U.msg) = struct
      type ustore = U.store
      let uget = T.hoist U.get
      let ugets viewer = T.hoist (U.gets viewer)
      let uput s = T.hoist (U.put s)
      let umodify modifier = T.hoist (U.modify modifier)
    end

    module Ref (U : REFBASE) (T : MONADT with type 'a ut := 'a U.t and type ulog := voidlog and type umsg := voidmsg) : REFBASE with type ref = U.ref and type value = U.value and type 'a t := 'a T.t = struct
      type ref = U.ref
      type value = U.value
      let newref v = T.hoist (U.newref v)
      let getref k = T.hoist (U.getref k)
      let putref k v = T.hoist (U.putref k v)
    end
    module RefW (U : sig include REFBASE type log end) (T : MONADT with type 'a ut := 'a U.t and type ulog := U.log and type umsg := voidmsg) : REFBASE with type ref = U.ref and type value = U.value and type 'a t := 'a T.t = struct
      type ref = U.ref
      type value = U.value
      let newref v = T.hoist (U.newref v)
      let getref k = T.hoist (U.getref k)
      let putref k v = T.hoist (U.putref k v)
    end
    module Ref_toW (U : REFBASE) (T : MONADTW with type 'a ut := 'a U.t and type umsg := voidmsg) : REFBASE with type ref = U.ref and type value = U.value and type 'a t := 'a T.t = struct
      type ref = U.ref
      type value = U.value
      let newref v = T.hoist (U.newref v)
      let getref k = T.hoist (U.getref k)
      let putref k v = T.hoist (U.putref k v)
    end
    module RefE (U : sig include REFBASE type msg end) (T : MONADT with type 'a ut := 'a U.t and type ulog := voidlog and type umsg := U.msg) : REFBASE with type ref = U.ref and type value = U.value and type 'a t := 'a T.t = struct
      type ref = U.ref
      type value = U.value
      let newref v = T.hoist (U.newref v)
      let getref k = T.hoist (U.getref k)
      let putref k v = T.hoist (U.putref k v)
    end
    module Ref_toE (U : REFBASE) (T : MONADTE with type 'a ut := 'a U.t and type ulog := voidlog) : REFBASE with type ref = U.ref and type value = U.value and type 'a t := 'a T.t = struct
      type ref = U.ref
      type value = U.value
      let newref v = T.hoist (U.newref v)
      let getref k = T.hoist (U.getref k)
      let putref k v = T.hoist (U.putref k v)
    end
    module RefW_toE (U : sig include REFBASE type log end) (T : MONADTE with type 'a ut := 'a U.t and type ulog := U.log) : REFBASE with type ref = U.ref and type value = U.value and type 'a t := 'a T.t = struct
      type ref = U.ref
      type value = U.value
      let newref v = T.hoist (U.newref v)
      let getref k = T.hoist (U.getref k)
      let putref k v = T.hoist (U.putref k v)
    end
    module RefE_toW (U : sig include REFBASE type msg end) (T : MONADTW with type 'a ut := 'a U.t and type umsg := U.msg) : REFBASE with type ref = U.ref and type value = U.value and type 'a t := 'a T.t = struct
      type ref = U.ref
      type value = U.value
      let newref v = T.hoist (U.newref v)
      let getref k = T.hoist (U.getref k)
      let putref k v = T.hoist (U.putref k v)
    end
    module RefX (U : sig include REFBASE type log type msg end) (T : MONADT with type 'a ut := 'a U.t and type ulog := U.log and type umsg := U.msg) : REFBASE with type ref = U.ref and type value = U.value and type 'a t := 'a T.t = struct
      type ref = U.ref
      type value = U.value
      let newref v = T.hoist (U.newref v)
      let getref k = T.hoist (U.getref k)
      let putref k v = T.hoist (U.putref k v)
    end

    module Ref_dbl (U : REFBASE) (T : REF_ with type 'a ut := 'a U.t and type ulog := voidlog and type umsg := voidmsg) = struct
      type uref = U.ref
      type uvalue = U.value
      let unewref v = T.hoist (U.newref v)
      let ugetref k = T.hoist (U.getref k)
      let uputref k v = T.hoist (U.putref k v)
    end
    module RefW_dbl (U : sig include REFBASE type log end) (T : REF_ with type 'a ut := 'a U.t and type ulog := U.log and type umsg := voidmsg) = struct
      type uref = U.ref
      type uvalue = U.value
      let unewref v = T.hoist (U.newref v)
      let ugetref k = T.hoist (U.getref k)
      let uputref k v = T.hoist (U.putref k v)
    end
    module RefE_dbl (U : sig include REFBASE type msg end) (T : REF_ with type 'a ut := 'a U.t and type ulog := voidlog and type umsg := U.msg) = struct
      type uref = U.ref
      type uvalue = U.value
      let unewref v = T.hoist (U.newref v)
      let ugetref k = T.hoist (U.getref k)
      let uputref k v = T.hoist (U.putref k v)
    end
    module RefX_dbl (U : sig include REFBASE type log type msg end) (T : REF_ with type 'a ut := 'a U.t and type ulog := U.log and type umsg := U.msg) = struct
      type uref = U.ref
      type uvalue = U.value
      let unewref v = T.hoist (U.newref v)
      let ugetref k = T.hoist (U.getref k)
      let uputref k v = T.hoist (U.putref k v)
    end

    module Writer (U : WRITERBASE) (T : MONADT with type 'a ut := 'a U.t and type ulog := U.log and type umsg := voidmsg) : WRITERBASE with type log = U.log and type 'a t := 'a T.t = struct
      type log = U.log
      let tell entries = T.hoist (U.tell entries)
      let listen xx = T.hoist_listen U.listen xx
      let pass xx = T.hoist_pass U.pass xx
      let listens selector xx = T.(>>=) (listen xx) (fun (x,w) -> T.mid (x, selector w))
      (* Obj.magic hack to make sufficiently polymorphic *)
      let censor c xx = Obj.magic (pass (T.(>>=) (Obj.magic xx) (fun x -> T.mid (x, c))))
    end
    module WriterE (U : sig include WRITERBASE type msg end) (T : MONADT with type 'a ut := 'a U.t and type ulog := U.log and type umsg := U.msg) : WRITERBASE with type log = U.log and type 'a t := 'a T.t = struct
      type log = U.log
      let tell entries = T.hoist (U.tell entries)
      let listen xx = T.hoist_listen U.listen xx
      let pass xx = T.hoist_pass U.pass xx
      let listens selector xx = T.(>>=) (listen xx) (fun (x,w) -> T.mid (x, selector w))
      (* Obj.magic hack to make sufficiently polymorphic *)
      let censor c xx = Obj.magic (pass (T.(>>=) (Obj.magic xx) (fun x -> T.mid (x, c))))
    end
    module Writer_toE (U : WRITERBASE) (T : MONADTE with type 'a ut := 'a U.t and type ulog := U.log) : WRITERBASE with type log = U.log and type 'a t := 'a T.t = struct
      type log = U.log
      let tell entries = T.hoist (U.tell entries)
      let listen xx = T.hoist_listen U.listen xx
      let pass xx = T.hoist_pass U.pass xx
      let listens selector xx = T.(>>=) (listen xx) (fun (x,w) -> T.mid (x, selector w))
      (* Obj.magic hack to make sufficiently polymorphic *)
      let censor c xx = Obj.magic (pass (T.(>>=) (Obj.magic xx) (fun x -> T.mid (x, c))))
    end
(*
    module Writer_toW (U : WRITERBASE) (T : MONADTW with type 'a ut := 'a U.t and type ulog := U.log and type umsg := voidmsg) : WRITERBASE with type log = U.log and type 'a t := 'a T.t = struct
      type log = U.log
      let tell entries = T.hoist (U.tell entries)
      let listen xx = T.hoist_listen U.listen xx
      let pass xx = T.hoist_pass U.pass xx
      let listens selector xx = T.(>>=) (listen xx) (fun (x,w) -> T.mid (x, selector w))
      (* Obj.magic hack to make sufficiently polymorphic *)
      let censor c xx = Obj.magic (pass (T.(>>=) (Obj.magic xx) (fun x -> T.mid (x, c))))
    end
    module WriterX (U : sig include WRITERBASE type msg end) (T : MONADT with type 'a ut := 'a U.t and type ulog := U.log and type umsg := U.msg) : WRITERBASE with type log = U.log and type 'a t := 'a T.t = struct
      type log = U.log
      let tell entries = T.hoist (U.tell entries)
      let listen xx = T.hoist_listen U.listen xx
      let pass xx = T.hoist_pass U.pass xx
      let listens selector xx = T.(>>=) (listen xx) (fun (x,w) -> T.mid (x, selector w))
      (* Obj.magic hack to make sufficiently polymorphic *)
      let censor c xx = Obj.magic (pass (T.(>>=) (Obj.magic xx) (fun x -> T.mid (x, c))))
    end
    module Writer_toW (U : WRITERBASE) (T : MONADT with type 'a ut := 'a U.t and type ulog := U.log and type umsg := voidmsg) : WRITERBASE with type log = U.log and type 'a t := 'a T.t = struct
      type log = U.log
      let tell entries = T.hoist (U.tell entries)
      let listen xx = T.hoist_listen U.listen xx
      let pass xx = T.hoist_pass U.pass xx
      let listens selector xx = T.(>>=) (listen xx) (fun (x,w) -> T.mid (x, selector w))
      (* Obj.magic hack to make sufficiently polymorphic *)
      let censor c xx = Obj.magic (pass (T.(>>=) (Obj.magic xx) (fun x -> T.mid (x, c))))
    end
    module WriterE_toW (U : sig include WRITERBASE type msg end) (T : MONADT with type 'a ut := 'a U.t and type ulog := U.log and type umsg := U.msg) : WRITERBASE with type log = U.log and type 'a t := 'a T.t = struct
      type log = U.log
      let tell entries = T.hoist (U.tell entries)
      let listen xx = T.hoist_listen U.listen xx
      let pass xx = T.hoist_pass U.pass xx
      let listens selector xx = T.(>>=) (listen xx) (fun (x,w) -> T.mid (x, selector w))
      (* Obj.magic hack to make sufficiently polymorphic *)
      let censor c xx = Obj.magic (pass (T.(>>=) (Obj.magic xx) (fun x -> T.mid (x, c))))
    end
*)

    module Error (U : ERRORBASE) (T : MONADT with type 'a ut := 'a U.t and type ulog := voidlog and type umsg := U.msg) : ERRORBASE with type msg = U.msg and type 'a error = 'a U.error and type 'a t := 'a T.t = struct
      type msg = U.msg
      type 'a error = 'a U.error
      let throw e = T.hoist (U.throw e)
      let catch xx handler = T.hoist_catch U.catch xx handler
    end
    module ErrorW (U : sig include ERRORBASE type log end) (T : MONADT with type 'a ut := 'a U.t and type ulog := U.log and type umsg := U.msg) : ERRORBASE with type msg = U.msg and type 'a error = 'a U.error and type 'a t := 'a T.t = struct
      type msg = U.msg
      type 'a error = 'a U.error
      let throw e = T.hoist (U.throw e)
      let catch xx handler = T.hoist_catch U.catch xx handler
    end
    module Error_toW (U : ERRORBASE) (T : MONADTW with type 'a ut := 'a U.t and type umsg := U.msg) : ERRORBASE with type msg = U.msg and type 'a error = 'a U.error and type 'a t := 'a T.t = struct
      type msg = U.msg
      type 'a error = 'a U.error
      let throw e = T.hoist (U.throw e)
      let catch xx handler = T.hoist_catch U.catch xx handler
    end
(*
    module Error_toE (U : ERRORBASE) (T : MONADTE with type 'a ut := 'a U.t and type ulog := voidlog) : ERRORBASE with type msg = U.msg and type 'a error = 'a U.error and type 'a t := 'a T.t = struct
      type msg = U.msg
      type 'a error = 'a U.error
      let throw e = T.hoist (U.throw e)
      let catch xx handler = T.hoist_catch U.catch xx handler
    end
    module ErrorX (U : sig include ERRORBASE type log end) (T : MONADT with type 'a ut := 'a U.t and type ulog := U.log and type umsg := U.msg) : ERRORBASE with type msg = U.msg and type 'a error = 'a U.error and type 'a t := 'a T.t = struct
      type msg = U.msg
      type 'a error = 'a U.error
      let throw e = T.hoist (U.throw e)
      let catch xx handler = T.hoist_catch U.catch xx handler
    end
    module Error_toE (U : ERRORBASE) (T : MONADT with type 'a ut := 'a U.t and type ulog := voidlog and type umsg := U.msg) : ERRORBASE with type msg = U.msg and type 'a error = 'a U.error and type 'a t := 'a T.t = struct
      type msg = U.msg
      type 'a error = 'a U.error
      let throw e = T.hoist (U.throw e)
      let catch xx handler = T.hoist_catch U.catch xx handler
    end
    module ErrorW_toE (U : sig include ERRORBASE type log end) (T : MONADT with type 'a ut := 'a U.t and type ulog := U.log and type umsg := U.msg) : ERRORBASE with type msg = U.msg and type 'a error = 'a U.error and type 'a t := 'a T.t = struct
      type msg = U.msg
      type 'a error = 'a U.error
      let throw e = T.hoist (U.throw e)
      let catch xx handler = T.hoist_catch U.catch xx handler
    end
*)

  end (* Derive *)

end (* Monad *)
