/*
 *  Yacc grammar for the parser. The files parser.mli and parser.ml
 *  are generated automatically from parser.mly.
 */

%{
open Lolevel.Error
open Interface
open Types
open Primitives

(* Used for de Bruijn-style binding; here a noop. *)
let addname identifier env = env

let create_literal fi l = fi, TermLiteral l

let create_prim fi prim = fun env -> fi, TermPrimfun prim

let create_cons fi cons = fun env -> fi, TermConstructor cons

let create_var tok = fun env -> tok.i, TermVar(tok.v)

let create_app1 head arg1 = fun env -> let head' = head env in fst head', TermApp(head', arg1 env)

let create_app2 head arg1 arg2 = fun env -> let head' = head env in fst head', TermApp((fst head', TermApp(head', arg1 env)), arg2 env)

let create_app3 head arg1 arg2 arg3 = fun env -> let head' = head env in fst head', TermApp((fst head',TermApp((fst head', TermApp(head', arg1 env)), arg2 env)), arg3 env)

let create_infix fi head arg1 arg2 = fun env -> let arg1' = arg1 env in fst arg1', TermApp((fst arg1', TermApp((fi, TermPrimfun head), arg1')), arg2 env)

let create_infix' fi head arg1 arg2 = fun env -> let arg2' = arg2 env in fst arg2', TermApp((fst arg2', TermApp((fi, TermPrimfun head), arg1 env)), arg2')

let create_or arg1 arg2 = fun env -> let arg1' = arg1 env in fst arg1', TermIf(arg1', create_literal void_fi (Bool true), arg2 env)

let create_and arg1 arg2 = fun env -> let arg1' = arg1 env in fst arg1', TermIf(arg1', arg2 env, create_literal void_fi (Bool false))

let create_abstracts (parms : (info * identifier) list) body = List.fold_left (fun prev (fi, ident) env -> fi, TermLambda(ident, prev (addname ident env))) body parms

let gensym = let nextsym = ref 0 in
             fun () -> let sym = !nextsym in
                       begin
                         incr nextsym;
                         "_" ^ string_of_int sym
                       end

let create_anon1 fi bodyfun = let ident = gensym () in fun env -> fi, TermLambda(ident, bodyfun (create_var {i=void_fi;v=ident}) (addname ident env))

let create_anon2 fi bodyfun = let ident1 = gensym () in let ident2 = gensym () in fun env -> fi, TermLambda(ident1, (fi, TermLambda(ident2, bodyfun (create_var {i=void_fi;v=ident1}) (create_var {i=void_fi;v=ident2}) (addname ident2 (addname ident1 env)))))

let create_anon3 fi bodyfun = let ident1 = gensym () in let ident2 = gensym () in let ident3 = gensym () in fun env -> fi, TermLambda(ident1, (fi, TermLambda(ident2, (fi, TermLambda(ident3, bodyfun (create_var {i=void_fi;v=ident1}) (create_var {i=void_fi;v=ident2}) (create_var {i=void_fi;v=ident3}) (addname ident3 (addname ident2 (addname ident1 env))))))))

%}

/* ---------------------------------------------------------------------- */
/* Preliminaries */

/* We first list all the tokens mentioned in the parsing rules
   below. The names of the tokens are common to the parser and the
   generated lexical analyzer. Each token is annotated with the type
   of data that it carries; normally, this is just file information
   (which is used by the parser to annotate the abstract syntax trees
   that it constructs), but sometimes -- in the case of identifiers
   and constant values -- more information is provided.
 */

%token <Lolevel.Error.info> EOF

/* Keyword tokens */
%token <Lolevel.Error.info> SYMBOL
%token <Lolevel.Error.info> IF
%token <Lolevel.Error.info> THEN
%token <Lolevel.Error.info> ELSE
%token <Lolevel.Error.info> LAMBDA
%token <Lolevel.Error.info> LET
%token <Lolevel.Error.info> LETREC
%token <Lolevel.Error.info> IN
%token <Lolevel.Error.info> END

%token <Lolevel.Error.info> MATCH
%token <Lolevel.Error.info> CASE
%token <Lolevel.Error.info> OF
%token <Lolevel.Error.info> WHEN
%token <Lolevel.Error.info> AS

%token <Lolevel.Error.info> TRUE
%token <Lolevel.Error.info> FALSE
%token <Lolevel.Error.info> SUCC
%token <Lolevel.Error.info> BOX
%token <Lolevel.Error.info> PRED
%token <Lolevel.Error.info> ISZERO
%token <Lolevel.Error.info> NOT
%token <Lolevel.Error.info> FST
%token <Lolevel.Error.info> SND
%token <Lolevel.Error.info> TRD

%token <Lolevel.Error.info> AND
%token <Lolevel.Error.info> OR
%token <Lolevel.Error.info> IDENTITY /* aka ICOMB */
%token <Lolevel.Error.info> COMP /* BCOMB is prefix */
%token <Lolevel.Error.info> PMOC
%token <Lolevel.Error.info> CONST /* aka KCOMB */
%token <Lolevel.Error.info> FLIP /* aka CCOMB */
%token <Lolevel.Error.info> SCOMB
%token <Lolevel.Error.info> BCOMB
%token <Lolevel.Error.info> TCOMB /* functional BOX */
%token <Lolevel.Error.info> VCOMB /* functional PAIR */
%token <Lolevel.Error.info> WCOMB
%token <Lolevel.Error.info> MCOMB /* aka omega */
%token <Lolevel.Error.info> LCOMB

/*
%token <Lolevel.Error.info> PAIR
%token <Lolevel.Error.info> TRIPLE
%token <Lolevel.Error.info> SWAP
%token <Lolevel.Error.info> HEAD
%token <Lolevel.Error.info> TAIL
%token <Lolevel.Error.info> CONS
%token <Lolevel.Error.info> NIL
*/

/* Identifier and constant value tokens */
%token <string Lolevel.Error.withinfo> UCID  /* uppercase-initial */
%token <string Lolevel.Error.withinfo> LCID  /* lowercase-initial and punctuation */
%token <int Lolevel.Error.withinfo> INTV
%token <int Lolevel.Error.withinfo> DOTV     /* 0. 1. and so on */
%token <string Lolevel.Error.withinfo> STRINGV

/* Symbolic tokens */
%token <Lolevel.Error.info> TILDE
%token <Lolevel.Error.info> BACKQUOTE
%token <Lolevel.Error.info> BANG
%token <Lolevel.Error.info> AT
%token <Lolevel.Error.info> HASH
%token <Lolevel.Error.info> DOLLAR
%token <Lolevel.Error.info> PERCENT
%token <Lolevel.Error.info> CARET
%token <Lolevel.Error.info> AMPER
%token <Lolevel.Error.info> STAR
%token <Lolevel.Error.info> LPAREN
%token <Lolevel.Error.info> RPAREN
%token <Lolevel.Error.info> UNDER
%token <Lolevel.Error.info> MINUS
%token <Lolevel.Error.info> PLUS
%token <Lolevel.Error.info> EQUAL
%token <Lolevel.Error.info> LCURLY
%token <Lolevel.Error.info> LSQUARE
%token <Lolevel.Error.info> RCURLY
%token <Lolevel.Error.info> RSQUARE
%token <Lolevel.Error.info> BAR
%token <Lolevel.Error.info> BACKSLASH
%token <Lolevel.Error.info> COLON
%token <Lolevel.Error.info> SEMI
%token <Lolevel.Error.info> DQUOTE
%token <Lolevel.Error.info> APOSTROPHE
%token <Lolevel.Error.info> LESS
%token <Lolevel.Error.info> COMMA
%token <Lolevel.Error.info> GREATER
%token <Lolevel.Error.info> DOT
%token <Lolevel.Error.info> QUESTION
%token <Lolevel.Error.info> SLASH

%token <Lolevel.Error.info> GREATEREQ
%token <Lolevel.Error.info> LESSEQ
%token <Lolevel.Error.info> EQUALEQ

/* For Church-encoded numbers, bools, tuples, lists */
%token <int Lolevel.Error.withinfo> HASHV    /* #0 #1 and so on */
%token <Lolevel.Error.info> HASHTRUE
%token <Lolevel.Error.info> HASHFALSE
%token <Lolevel.Error.info> HASHLPAREN
%token <Lolevel.Error.info> HASHLSQUARE

%nonassoc IN
%nonassoc LET LETREC /* OCaml puts above SEMI for sequencing */
%nonassoc OF /* below SEMI */
%nonassoc THEN /* below ELSE */
%nonassoc ELSE
%nonassoc AS
%left BAR SEMI /* below =/match */
%nonassoc DOT
%left COMMA
/* Associativity and precedence for infix operators */
%right DOLLAR               /* lowest precedence */
/* >>,>>= are left1; flipped are right1 */
%right OR
%right AND
/* TODO handle sequences? */
%nonassoc LESS GREATER LESSEQ GREATEREQ EQUALEQ
/* right CONS */
%left MINUS PLUS
%left STAR
/* right exponentiation */
%left PMOC
%right COMP              /* tightest precedence */
/* constructor application, above AS BAR CONS COMMA */
%nonassoc FLIP CONSTRUCTOR SUCC BOX
%nonassoc ATOMIC IDENTITY CONST PRED ISZERO NOT FST SND TRD TRUE FALSE INTV DOTV HASHV STRINGV LCID UCID LPAREN LSQUARE LCURLY SCOMB BCOMB TCOMB VCOMB WCOMB MCOMB LCOMB HASHLPAREN HASHLSQUARE HASHTRUE HASHFALSE
%nonassoc INLINE

/*
BEGIN
SYMBOL
IF
LAMBDA
LETREC
END
MATCH
CASE
WHEN

TILDE
BACKQUOTE
BANG
AT
HASH
PERCENT
CARET
AMPER
RPAREN
UNDER
EQUAL
RCURLY
RSQUARE
BACKSLASH
COLON
DQUOTE
APOSTROPHE
QUESTION
SLASH
*/

/* ---------------------------------------------------------------------- */
/* The starting production of the generated parser is the
   syntactic class toplevel. The type that is returned when a
   toplevel is recognized is
     Types.env -> (Types.toplevel_syntax list * Types.env)
   that is, the parser returns to the user program a function
   that, when given a naming context, returns a fully parsed
   list of toplevel_syntax and the new naming context that
   results when all the names bound in these toplevels are
   defined.

   All of the syntactic productions in the parser follow the
   same pattern: they take a context as argument and return a
   fully parsed abstract syntax tree (and, if they involve any
   constructs that bind variables in some following phrase, a
   new context).
*/

%start total
%type < Types.env -> (Types.toplevel_syntax list * Types.env) > total
%%

/* ---------------------------------------------------------------------- */
/* Main body of the parser definition */

/* The outermost production of a file is a sequence of Toplevels, separated by semicolons. */
total :
  | EOF
      { fun env -> [],env }
  | Toplevel EOF
      { fun env ->
          let (top,env) = $1 env in
          top [],env }
  | Toplevel SEMI total
      { fun env ->
          let (top,env') = $1 env in
          let (tops,env'') = $3 env' in
          top tops, env'' }
/* | SEMI total { $2 } */


Toplevel :
  | Term
      { fun env -> (let t = $1 env in fun tops -> Term(fst t,t)::tops), env }
  | SYMBOL LCID
      { fun env -> (fun tops -> BindSymbol($2.i,$2.v)::tops), addname $2.v env }
  | LET LetBindings END
      { fun env ->
          let (binds, env') = List.fold_left
                                (fun (bs,env) (fi,ident,arg) ->
                                   BindTerm(fi,ident,arg env)::bs, addname ident env)
                                ([], env) $2 in
          (fun tops -> List.append binds tops), env' }
  /* we permit optional trailing SEMI */
  | LET LetBindings SEMI END
      { fun env ->
          let (binds, env') = List.fold_left
                                (fun (bs,env) (fi,ident,arg) ->
                                   BindTerm(fi,ident,arg env)::bs, addname ident env)
                                ([], env) $2 in
          (fun tops -> List.append binds tops), env' }



LambdaBindings :
  | LAMBDA LCID { [($1 (* $2.i *),$2.v)] }
  | LAMBDA UNDER { [($1 (* $2 *),"_")] }
  | BACKSLASH LCID { [($1 (* $2.i *),$2.v)] }
  | BACKSLASH UNDER { [($1 (* $2 *),"_")] }
  | LambdaBindings LCID { ($2.i,$2.v)::$1 }
  | LambdaBindings UNDER { ($2,"_")::$1 }

FunctDeclaration :
  | LCID LCID { $1, [($2.i,$2.v)] }
  | LCID UNDER { $1, [($2,"_")] }
  | FunctDeclaration LCID { let (head,parms) = $1 in head, ($2.i,$2.v)::parms }
  | FunctDeclaration UNDER { let (head,parms) = $1 in head, ($2,"_")::parms }

LetBindings :
  | LCID EQUAL Term { [($1.i,$1.v,$3)] }
  | UNDER EQUAL Term { [($1,"_",$3)] }
  | FunctDeclaration EQUAL Term
      { let (head,parms) = $1 in [(head.i, head.v, create_abstracts parms $3)] }
  | LetBindings SEMI LCID EQUAL Term { ($3.i,$3.v,$5)::$1 }
  | LetBindings SEMI UNDER EQUAL Term { ($3,"_",$5)::$1 }
  | LetBindings SEMI FunctDeclaration EQUAL Term
      { let (head,parms) = $3 in (head.i, head.v, create_abstracts parms $5)::$1 }

Term :
  | InfixTerm
      { snd $1 }
  | IF Term THEN Term ELSE Term
      { fun env -> $1, TermIf($2 env, $4 env, $6 env) }
  | LambdaBindings DOT Term
      { create_abstracts $1 $3 }
  | LET LetBindings IN Term
      { fun env -> List.fold_left
                     (fun prev (fi,ident,arg) env ->
                        fi, TermLet(ident, arg env, prev (addname ident env)))
                   $4 $2 env }
  /* we permit optional trailing SEMI */
  | LET LetBindings SEMI IN Term
      { fun env -> List.fold_left
                     (fun prev (fi,ident,arg) env ->
                        fi, TermLet(ident, arg env, prev (addname ident env)))
                   $5 $2 env }
  | LETREC LCID EQUAL Term IN Term
      { fun env -> $1, TermLetRec([$2.v, $4 env], $6 (addname $2.v env)) }
  | LETREC UNDER EQUAL Term IN Term
      { fun env -> $1, TermLetRec(["_", $4 env], $6 (addname "_" env)) }
/*
  | error { print_endline "error"; fun env -> void_fi, TermLiteral Unit }
*/

/* These return a pair of (precedence, term_value) */
InfixTerm :
  | AppTerm { 11, $1 }
  | InfixTerm DOLLAR InfixTerm    { 0, create_app1 (snd $1) (snd $3) }
  | InfixTerm OR InfixTerm        { 2, create_or (snd $1) (snd $3) }
  | InfixTerm AND InfixTerm       { 3, create_and (snd $1) (snd $3) }
  /* TODO handle sequences? */
  | InfixTerm LESS InfixTerm      { 4, create_infix $2 (_Lt) (snd $1) (snd $3) }
  | InfixTerm GREATER InfixTerm   { 4, create_infix' $2 (_Lt) (snd $3) (snd $1) }
  | InfixTerm LESSEQ InfixTerm    { 4, create_infix $2 (_Leq) (snd $1) (snd $3) }
  | InfixTerm GREATEREQ InfixTerm { 4, create_infix' $2 (_Leq) (snd $3) (snd $1) }
  | InfixTerm EQUALEQ InfixTerm   { 4, create_infix $2 (_Eq) (snd $1) (snd $3) }
  | InfixTerm MINUS InfixTerm     { 6, create_infix $2 (_Minus) (snd $1) (snd $3) }
  | InfixTerm PLUS InfixTerm      { 6, create_infix $2 (_Plus) (snd $1) (snd $3) }
  | InfixTerm STAR InfixTerm      { 7, create_infix $2 (_Times) (snd $1) (snd $3) }
  | InfixTerm PMOC InfixTerm      { 9, create_anon1 $2 (fun ident -> create_app1 (snd $3) (create_app1 (snd $1) ident)) }
  | InfixTerm COMP InfixTerm      { 10, create_anon1 $2 (fun ident -> create_app1 (snd $1) (create_app1 (snd $3) ident)) }

InfixOp :
  | OR        { 2, create_anon2 $1 create_or }
  | AND       { 3, create_anon2 $1 create_and }
  | LESS      { 4, create_prim $1 (_Lt) }
  | GREATER   { 4, create_anon2 $1 (fun ident1 ident2 -> create_app2 (create_prim $1 (_Lt)) ident2 ident1) }
  | LESSEQ    { 4, create_prim $1 (_Leq) }
  | GREATEREQ { 4, create_anon2 $1 (fun ident1 ident2 -> create_app2 (create_prim $1 (_Leq)) ident2 ident1) }
  | EQUALEQ   { 4, create_prim $1 (_Eq) }
  | MINUS     { 6, create_prim $1 (_Minus) }
  | PLUS      { 6, create_prim $1 (_Plus) }
  | STAR      { 7, create_prim $1 (_Times) }
  | PMOC      { 9, create_anon3 $1 (fun ident1 ident2 ident3 -> create_app1 ident2 (create_app1 ident1 ident3)) }
  | COMP      { 10, create_anon3 $1 (fun ident1 ident2 ident3 -> create_app1 ident1 (create_app1 ident2 ident3)) }

FlippedInfixOp :
  | OR        { 2, create_anon2 $1 (fun ident1 ident2 -> create_or ident2 ident1) }
  | AND       { 3, create_anon2 $1 (fun ident1 ident2 -> create_and ident2 ident1) }
  | LESS      { 4, create_anon2 $1 (fun ident1 ident2 -> create_app2 (create_prim $1 (_Lt)) ident2 ident1) }
  | GREATER   { 4, create_prim $1 (_Lt) }
  | LESSEQ    { 4, create_anon2 $1 (fun ident1 ident2 -> create_app2 (create_prim $1 (_Leq)) ident2 ident1) }
  | GREATEREQ { 4, create_prim $1 (_Leq) }
  | EQUALEQ   { 4, create_prim $1 (_Eq) }
  | MINUS     { 6, create_anon2 $1 (fun ident1 ident2 -> create_app2 (create_prim $1 (_Minus)) ident2 ident1) }
  | PLUS      { 6, create_prim $1 (_Plus) }
  | STAR      { 7, create_prim $1 (_Times) }
  | PMOC      { 9, create_anon3 $1 (fun ident1 ident2 ident3 -> create_app1 ident1 (create_app1 ident2 ident3)) }
  | COMP      { 10, create_anon3 $1 (fun ident1 ident2 ident3 -> create_app1 ident2 (create_app1 ident1 ident3)) }

Section : /* also Pairs and Triples */
  | LPAREN COMMA RPAREN             { create_cons $1 (_Pair) }
  | LPAREN InfixTerm COMMA RPAREN       { create_app1 (create_cons $1 (_Pair)) (snd $2) }
  | LPAREN COMMA InfixTerm RPAREN       { create_anon1 $1 (fun ident -> create_app2 (create_cons $1 (_Pair)) ident (snd $3)) }
  | LPAREN InfixTerm COMMA InfixTerm RPAREN { create_app2 (create_cons $1 (_Pair)) (snd $2) (snd $4) }
  | LPAREN COMMA COMMA RPAREN       { create_cons $1 (_Triple) }
  | LPAREN InfixTerm COMMA COMMA RPAREN { create_app1 (create_cons $1 (_Triple)) (snd $2) }
  | LPAREN COMMA InfixTerm COMMA RPAREN { create_anon1 $1 (fun ident -> create_app2 (create_cons $1 (_Triple)) ident (snd $3)) }
  | LPAREN COMMA COMMA InfixTerm RPAREN { create_anon2 $1 (fun ident1 ident2 -> create_app3 (create_cons $1 (_Triple)) ident1 ident2 (snd $4)) }
  | LPAREN COMMA InfixTerm COMMA InfixTerm RPAREN { create_anon1 $1 (fun ident -> create_app3 (create_cons $1 (_Triple)) ident (snd $3) (snd $5)) }
  | LPAREN InfixTerm COMMA COMMA InfixTerm RPAREN { create_anon1 $1 (fun ident -> create_app3 (create_cons $1 (_Triple)) (snd $2) ident (snd $5)) }
  | LPAREN InfixTerm COMMA InfixTerm COMMA RPAREN { create_app2 (create_cons $1 (_Triple)) (snd $2) (snd $4) }
  | LPAREN InfixTerm COMMA InfixTerm COMMA InfixTerm RPAREN { create_app3 (create_cons $1 (_Triple)) (snd $2) (snd $4) (snd $6) }
  | LPAREN DOLLAR RPAREN            { create_anon2 $1 create_app1 }
  | LPAREN DOLLAR InfixTerm RPAREN  { create_anon1 $1 (fun ident -> create_app1 ident (snd $3)) }
  | LPAREN InfixTerm DOLLAR RPAREN  { create_anon1 $1 (fun ident -> create_app1 (snd $2) ident) (* eta-expanded $2, may get stuck differently *) }
  | LPAREN InfixOp RPAREN           { (snd $2) }
  | LPAREN InfixOp InfixTerm RPAREN { if fst $3 < fst $2 then raise Parsing.Parse_error else create_anon1 $1 (fun ident -> create_app2 (snd $2) ident (snd $3)) }
  | LPAREN InfixTerm InfixOp RPAREN { (* if fst $2 < fst $3 then raise Parsing.Parse_error else *) create_app1 (snd $3) (snd $2) }

Constructor1 : /* excluding Pairs and Triples */
  | SUCC
      { $1, create_cons $1 (_Succ) }
  | BOX
      { $1, create_cons $1 (_Box) }

AppTerm :
  | AtomicTerm
      { $1 }
  | AppTerm AtomicTerm
      { create_app1 $1 $2 }
  | Constructor1 AtomicTerm %prec INLINE /* in case we forbid unapplied constructors */
      { create_app1 (snd $1) $2 }
  | FLIP LPAREN COMMA RPAREN %prec INLINE
      { create_anon2 $1 (fun ident1 ident2 -> create_app2 (create_cons $2 (_Pair)) ident2 ident1) }
  /* TODO FLIP (InfixTerm,,) or (,InfixTerm,) or (,,InfixTerm) */
  | FLIP LPAREN DOLLAR RPAREN %prec INLINE /* same as TCOMB */
      { create_anon2 $1 (fun ident1 ident2 -> create_app1 ident2 ident1) }
  | FLIP LPAREN FlippedInfixOp RPAREN %prec INLINE
      { (snd $3) }
  | FLIP LPAREN Term RPAREN %prec FLIP
      { create_anon2 $1 (fun ident1 ident2 -> create_app2 $3 ident2 ident1) }

/* Atomic terms are ones that never require extra parentheses */
AtomicTerm :
  | LPAREN RPAREN %prec ATOMIC
      { create_cons $1 (_Unit) }
  | Section %prec ATOMIC
      { $1 }
  | LPAREN Term RPAREN %prec ATOMIC
      { $2 }
  | Constructor1 %prec CONSTRUCTOR
      {
        (* create_anon1 (fst $1) (fun ident -> create_app1 (snd $1) ident) *)
        snd $1
      }
  | IDENTITY
      { create_anon1 $1 (fun ident -> ident) }
  | CONST
      { create_anon2 $1 (fun ident1 ident2 -> ident1) }
  | FLIP
      { create_anon3 $1 (fun ident1 ident2 ident3 -> create_app2 ident1 ident3 ident2) }
  | PRED
      { fun env -> $1,TermDestructor Pred }
  | ISZERO
      { fun env -> $1,TermDestructor IsZero }
  | NOT
      { create_prim $1 (_Not) }
  | FST
      { fun env -> $1,TermDestructor Fst }
  | SND
      { fun env -> $1,TermDestructor Snd }
  | TRD
      { fun env -> $1,TermDestructor Trd }
  | TRUE
      { fun env -> create_literal $1 (Bool true) }
  | FALSE
      { fun env -> create_literal $1 (Bool false) }
  | INTV
      { fun env -> create_literal $1.i (Num $1.v) }
  | DOTV
      { fun env ->
          let rec aux = function
            | 0 -> create_cons $1.i _Zero env
            | n -> $1.i, TermApp(create_cons $1.i (_Succ) env, aux (n-1)) in
          aux $1.v }
  | STRINGV
      { fun env -> create_literal $1.i (Str $1.v) }
  | LCID
      { create_var $1 }
  | SCOMB
      { create_anon3 $1 (fun ident1 ident2 ident3 -> create_app2 ident1 ident3 (create_app1 ident2 ident3)) }
  | BCOMB /* same as (COMP) */
      { create_anon3 $1 (fun ident1 ident2 ident3 -> create_app1 ident1 (create_app1 ident2 ident3)) }
  | TCOMB /* same as flip ($) */
      { create_anon2 $1 (fun ident1 ident2 -> create_app1 ident2 ident1) }
  | VCOMB
      { create_anon3 $1 (fun ident1 ident2 ident3 -> create_app2 ident3 ident1 ident2) }
  | WCOMB
      { create_anon2 $1 (fun ident1 ident2 -> create_app2 ident1 ident2 ident2) }
  | MCOMB
      { create_anon1 $1 (fun ident1 -> create_app1 ident1 ident1) }
  | LCOMB
      { create_anon2 $1 (fun ident1 ident2 -> create_app1 ident1 (create_app1 ident2 ident2)) }

/*   */
